import copy
import logging
import os
import yaml

from typing import Dict

from dataspree.inference.core.config.config_store import ConfigStore

logger = logging.getLogger(__name__)


class LocalStore(ConfigStore):
    def __init__(self, store_path) -> None:
        self._store_path = store_path
        self._data: Dict = dict()

    def load_data(self):
        self._load_data()

    def _load_data(self):
        try:
            with open(self._store_path, encoding='utf8') as f:
                _data = yaml.safe_load(f)
                if not isinstance(_data, dict):
                    raise Exception('File must not be parsed as raw string.')
                self._data = _data

        except FileNotFoundError:
            self._data = dict()
            self.save_data()

        except Exception as e:
            logger.error(f'Could not load data from {self._store_path}. '
                         f'Override with default. Exception: {e}')

            # Backup corrupted configuration file.
            file, extension = os.path.splitext(self._store_path)
            os.replace(self._store_path, f'{file}-backup{extension}')

            # Reset Configuration and clear configuration file.
            self._data = dict()
            self.save_data()

    def empty(self):
        return len(self._data) == 0

    def save_raw_and_reload(self, raw_data):
        with open(self._store_path, 'w') as f:
            f.write(raw_data)
        self._load_data()

    def save_data(self):
        self._save_data()

    def _save_data(self):
        with open(self._store_path, 'w') as f:
            f.write(yaml.dump(self._data))

    def get(self, name, default=None):
        return self._data.get(name, default)

    def get_data(self):
        return copy.deepcopy(self._data)

    def put(self, name, data):
        self._data[name] = data

    def put_data(self, data):
        self._data = data
