# mypy: ignore-errors

from functools import reduce
from typing import Optional, List, Tuple, Dict, Union, Any

import numpy as np
import scipy.optimize as sp
import copy

from dataspree.inference.core.environment import is_application_debug


class OptimalAssignment:

    @staticmethod
    def get_cost(cost_matrix: np.ndarray, assignment: Tuple[np.ndarray, np.ndarray]) -> float:
        return np.sum(cost_matrix[assignment])

    @staticmethod
    def optimize(cost_matrix: np.ndarray, number_of_elements: int = 1):
        ...


class Hungarian(OptimalAssignment):

    @staticmethod
    def optimize(cost_matrix: np.ndarray, number_of_elements: int = 1) \
            -> Optional[Tuple[np.ndarray, np.ndarray]]:
        assert number_of_elements == 1
        try:
            return sp.linear_sum_assignment(cost_matrix, maximize=False)
        except Exception as e:
            return None


class Murphy(OptimalAssignment):

    @staticmethod
    def optimize(cost_matrix: np.ndarray, number_of_elements: int = 1, return_cost: bool = False) \
            -> Optional[Union[Tuple[List[Tuple[np.ndarray, np.ndarray]], List[float]],
                                    List[Tuple[np.ndarray, np.ndarray]]]]:
        cost_matrix = cost_matrix.copy()
        og_cost = cost_matrix.copy()

        number_of_rows: int
        number_of_cols: int
        number_of_rows, number_of_cols = cost_matrix.shape

        # insert first entry to the list of nodes
        first_assignment: Optional[Tuple[np.ndarray, np.ndarray]] = Hungarian.optimize(cost_matrix=cost_matrix)
        if first_assignment is None:
            return None

        prior_assignment_map_rows: List[int] = list(range(number_of_rows))
        prior_assignment_map_cols: List[int] = list(range(number_of_cols))
        first_cost = OptimalAssignment.get_cost(cost_matrix=cost_matrix, assignment=first_assignment)

        nodes: List[Tuple[np.ndarray, List[int], List[int], Dict[int, int], float, Tuple[np.ndarray, np.ndarray]]] \
            = [(cost_matrix, prior_assignment_map_rows, prior_assignment_map_cols, {}, first_cost, first_assignment)]

        assignments: List[Tuple[np.ndarray, np.ndarray]] = []
        costs: List[float] = []

        for i in range(1, number_of_elements + 1):  # XXX:
            if not len(nodes):
                break

            next = Murphy.next(nodes, og_cost, last_iteration=i >= number_of_elements)
            if next is None:
                break

            ass, cost = next

            assignments += [ass]
            costs += [cost]

        if return_cost:
            return assignments, costs

        return assignments

    @staticmethod
    def next(nodes: List[Tuple[np.ndarray, List[int], List[int], Dict[int, int], float, Tuple[np.ndarray, np.ndarray]]],
             og_cost: np.ndarray, last_iteration: bool = False) -> Optional[
        Tuple[Tuple[np.ndarray, np.ndarray], float]]:

        best_assignment = Murphy.get_best_assignment(nodes)

        if best_assignment is None:
            return None

        current_node: Tuple[np.ndarray, List[int], List[int], Dict[int, int], float, Tuple[np.ndarray, np.ndarray]] \
            = nodes.pop(best_assignment)

        # Cost matrix
        current_cost_matrix: np.ndarray

        # Information required for transforming solution of the submatrix' cost minimization problem to a full
        # assignment
        prior_assignment_map_rows: List[int]  #: small cost matrix index -> large cost matrix index
        prior_assignment_map_cols: List[int]
        prior_assignment_map_vals: Dict[int, int]

        # Best assignment of current_cost_matrix
        sub_cost: float
        sub_assignment: Tuple[np.ndarray, np.ndarray]

        current_cost_matrix, prior_assignment_map_rows, prior_assignment_map_cols, \
            prior_assignment_map_vals, sub_cost, sub_assignment = current_node

        # assemble full assignment
        mat_sz: int = len(prior_assignment_map_vals) + len(sub_assignment[0])
        full_assignment: Tuple[np.ndarray, np.ndarray] = (
            np.empty([mat_sz], dtype=int), np.empty([mat_sz], dtype=int))

        for i, (row, col) in enumerate(prior_assignment_map_vals.items()):
            full_assignment[0][i] = row
            full_assignment[1][i] = col

        for j in range(len(sub_assignment[0])):
            row = prior_assignment_map_rows[sub_assignment[0][j]]
            col = prior_assignment_map_cols[sub_assignment[1][j]]

            full_assignment[0][len(prior_assignment_map_vals) + j] = row
            full_assignment[1][len(prior_assignment_map_vals) + j] = col
        assert len(prior_assignment_map_vals) + len(sub_assignment[0]) == len(full_assignment[0])

        if not last_iteration:
            running_cost: float = np.sum(og_cost[list(prior_assignment_map_vals.keys()),
            list(prior_assignment_map_vals.values())])

            # Populate matrix with new nodes
            npa = np.array(sub_assignment).T
            for i, (track_index_og, detection_index_og) in enumerate(npa[:-1]):
                track_index = track_index_og - np.sum(npa[:i, 0] < track_index_og)
                detection_index = detection_index_og - np.sum(npa[:i, 1] < detection_index_og)

                current_cost_matrix[track_index, detection_index] = np.inf

                assignment = Hungarian.optimize(cost_matrix=current_cost_matrix)

                if assignment is not None:
                    cost: float = running_cost + OptimalAssignment.get_cost(current_cost_matrix, assignment=assignment)

                    assert len(assignment[0]) == len(current_cost_matrix)
                    assert len(current_cost_matrix)

                    nodes.append((current_cost_matrix,
                                  copy.copy(prior_assignment_map_rows),
                                  copy.copy(prior_assignment_map_cols),
                                  copy.copy(prior_assignment_map_vals),
                                  cost, assignment))

                current_cost_matrix = np.delete(np.delete(current_cost_matrix, track_index, axis=0),
                                                detection_index, axis=1)

                # Update the prior assignment map
                # prior_assignment_map_vals[prior_assignment_map_rows[track_index_og]] \
                #    = prior_assignment_map_cols[detection_index_og]

                og_track, og_det = prior_assignment_map_rows[track_index], prior_assignment_map_cols[detection_index]

                prior_assignment_map_vals[og_track] = og_det
                running_cost += og_cost[og_track, og_det]

                prior_assignment_map_rows.pop(track_index)
                prior_assignment_map_cols.pop(detection_index)

                ...

        return full_assignment, sub_cost

    @staticmethod
    def get_best_assignment(nodes: List[Tuple[Any, Any, Any, Any, float, Any]]) -> int:
        # Find the best node in the set of all nodes
        index_best_node: int = -1
        cost_best_node = np.inf
        for i, node in enumerate(nodes):
            c_cost: float = node[4]
            if c_cost <= cost_best_node:
                cost_best_node = c_cost
                index_best_node = i

        return index_best_node

    @staticmethod
    def optimize_incorrect(cost_matrix: np.ndarray, number_of_elements: int = 1, return_cost: bool = False):
        # XXX: incorrect

        current_cost_matrix = cost_matrix.copy()

        previous_best_assignment = Hungarian.optimize(cost_matrix=current_cost_matrix)
        if previous_best_assignment is None:
            return

        best_assignments = [previous_best_assignment]
        best_costs = [OptimalAssignment.get_cost(cost_matrix, previous_best_assignment)]

        for n in range(1, number_of_elements):

            if previous_best_assignment is None:
                return

            current_cost_matrices = np.repeat(current_cost_matrix[None], len(previous_best_assignment[0]), axis=0)

            best_cost: float = np.inf
            best_index: int = -1
            best_assignment: Optional[Tuple[np.ndarray, np.ndarray]] = None

            for i, (track_index, detection_index) in enumerate(np.array(previous_best_assignment).T):

                current_cost_matrices[i, track_index, detection_index] = np.inf

                # Here we could copy the matrix and make the assignment problem a bit easier
                assignment = Hungarian.optimize(cost_matrix=current_cost_matrices[i])
                if assignment is None:
                    return
                cost: float = OptimalAssignment.get_cost(cost_matrix=cost_matrix, assignment=assignment)

                if cost <= best_cost:
                    best_index = i
                    best_cost = cost
                    best_assignment = assignment

            current_cost_matrix = current_cost_matrices[best_index]
            previous_best_assignment = best_assignment

            if previous_best_assignment is None:
                return

            best_assignments += [previous_best_assignment]
            best_costs += [best_cost]


        if return_cost:
            return best_assignments, best_costs
        return best_assignments


if __name__ == '__main__':
    test_matrix = np.array([[7, 51, 52, 87, 38, 60, 74, 66, 0, 20],
                            [50, 12, 0, 64, 8, 53, 0, 46, 76, 42],
                            [27, 77, 0, 18, 22, 48, 44, 13, 0, 57],
                            [62, 0, 3, 8, 5, 6, 14, 0, 26, 39],
                            [0, 97, 0, 5, 13, 0, 41, 31, 62, 48],
                            [79, 68, 0, 0, 15, 12, 17, 47, 35, 42],
                            [76, 99, 48, 27, 34, 0, 0, 0, 28, 0],
                            [0, 20, 9, 27, 46, 15, 84, 19, 3, 24],
                            [56, 10, 45, 39, 0, 93, 67, 79, 19, 38],
                            [27, 0, 39, 53, 46, 24, 69, 46, 23, 1]]).astype(np.float64)

    # resulting_optimal_assignments, resulting_costs = Murphy.optimize(test_matrix.copy(), 1000, return_cost=True)

    test_matrix = test_matrix[:5, :5]
    n_vals = reduce(lambda a, b: a * b, range(1, 6)) + 1
    opt = Murphy.optimize(test_matrix.copy(), n_vals + 1, return_cost=True)

    if opt is None:
        raise Exception('opt is None')

    resulting_optimal_assignments, resulting_costs = opt

    """
    for i, (a, c) in enumerate(zip(resulting_optimal_assignments, resulting_costs)):
        if __debug__ and is_application_debug():
            assert all((i in a[0] and i in a[1] for i in range(test_matrix.shape[0])))

        ccc = OptimalAssignment.get_cost(cost_matrix=test_matrix, assignment=a)
        if __debug__ and is_application_debug():
            assert ccc == c

        print(i + 1, n_vals, [aa.astype(int) + 1 for aa in a], c)

    for i, (a, c) in enumerate(zip(resulting_optimal_assignments, resulting_costs)):
        for j, (aj, cj) in enumerate(zip(resulting_optimal_assignments, resulting_costs)):
            if j <= i:
                continue

            if __debug__ and is_application_debug():
                assert cj >= c
                assert np.all(a[0] == aj[0]) and np.any(a[1] != aj[1])
    """
