import copy
from abc import ABCMeta

from dataclasses import dataclass, field
from typing import Dict, Optional, List, Set, Union, Any
from enum import Enum
import numpy as np

from dataspree.inference.sensor.data.camera_model import CameraModel
from dataspree.inference.sensor.data.exceptions import SensorException
from dataspree.inference.sensor.data.pixel_format import PixelFormat, GenicamPixelFormat
from dataspree.inference.sensor.data.sensor import (SensorData, Sensor, SensorState, OptionalConfiguration,
                                                    NotAvailable, NotRetrieved, is_available)
from dataspree.inference.sensor.data.sensor_model import SensorModelDescriptor


class TriggerType(Enum):
    """Offers distinction between software and hardware (and no) trigger."""
    SOFTWARE = 'software'
    HARDWARE = 'hardware'


@dataclass
class TriggerMode(OptionalConfiguration):
    """Contains ways of triggering

    Args:
        trigger_source (str) descriptor of the source of the trigger.

        trigger_delay (Union[int, NotAvailable]) Delay between the trigger and the action it causes.
    """
    trigger_source: str
    trigger_delay: Union[int, NotAvailable]


@dataclass
class SoftwareTrigger(TriggerMode):
    """Software trigger

    Inherits from TriggerMode and represents a software-based trigger mechanism.

    Attributes:
        trigger_source (str): Descriptor of the source of the trigger.
        trigger_delay (Union[int, NotAvailable]): Delay between the trigger and the action it causes.
    """


class TriggerActivation(Enum):
    """Contains different trigger transitions that can can either express an event time or duration."""
    RISING_EDGE = 'on rising edge'
    FALLING_EDGE = 'on falling edge'
    ANY_EDGE = 'on any edge'
    LEVEL_HIGH = 'while high'
    LEVEL_LOW = 'while low'


@dataclass
class HardwareTrigger(TriggerMode):
    """Hardware trigger

    Args:
        trigger_activation (TriggerActivation): The transition that is supposed to activate the trigger.
    """
    trigger_activation: TriggerActivation


@dataclass
class AcquisitionMode(OptionalConfiguration):
    """Determines the way in which the sensor acquires data.

    Attributes:
        max_rate: Union[float, NotAvailable] The maximal frame or line rate [Hz], that the sensor should never exceed.
    """
    max_rate: Union[float, NotAvailable]


@dataclass
class ContinuousAcquisitionMode(AcquisitionMode):
    """The sensor automatically acquires data, and does not need to rely on a hardware or software trigger. """

    def __str__(self):
        return f'Continuous{f"@{self.max_rate:0.2f}" if is_available(self.max_rate) else str()}'


class TemperatureState(Enum):
    """Contains different temperature states sensors can have."""
    OKAY = 'okay'
    CRITICAL = 'critical'
    ERROR = 'beyond critical'


@dataclass
class TriggerAcquisitionMode(AcquisitionMode):
    """The sensor acquires data when triggered.

    Attributes:

        trigger_start (TriggerMode) the frame / line start trigger.

        trigger_frame_end (TriggerMode) trigger, that indicates the end of the frame.

        trigger_frame_burst_start (TriggerMode) trigger to start frame burst.

        trigger_frame_burst_end (TriggerMode) trigger to stop frame burst.

        burst_frame_count (Union[int, NotAvailable]) the frame count to be acquired in one burst.

        burst_mode (Union[str, NotAvailable]) the acquisition burst mode.

    """
    trigger_start: TriggerMode
    trigger_frame_end: Union[TriggerMode, NotAvailable, None] = field(default_factory=NotAvailable, init=True)

    trigger_frame_burst_start: Union[TriggerMode, NotAvailable, None] = field(default_factory=NotAvailable, init=True)
    trigger_frame_burst_end: Union[TriggerMode, int, NotAvailable, None] = field(default_factory=NotAvailable,
                                                                                 init=True)
    burst_frame_count: Union[int, NotAvailable, None] = field(default_factory=NotAvailable, init=True)
    burst_mode: Union[str, NotAvailable, None] = field(default_factory=NotAvailable, init=True)

    def __str__(self) -> str:
        return (f'Trigger {self.trigger_start.trigger_source or str()},'
                f'{self.trigger_frame_end if is_available(self.trigger_frame_end) else "/"} '
                f'Burst {self.trigger_frame_burst_start if is_available(self.trigger_frame_burst_start) else "/"},'
                f'{self.trigger_frame_burst_end if is_available(self.trigger_frame_burst_end) else "/"},'
                f' {self.burst_frame_count if is_available(self.burst_frame_count) else str()},'
                f'{self.burst_mode if is_available(self.burst_mode) else str()}.')


@dataclass
class Roi(OptionalConfiguration):
    """Region of interest.
    Arguments:
        x (int): x position in pixel.
        y (int): y position in pixel.
        width (int): with in pixel.
        height (int): height in pixel.
    """
    x: int
    y: int
    width: int
    height: int


@dataclass
class TonalRange(OptionalConfiguration):
    """Represents the tonal range configuration for a camera.

    Inherits from OptionalConfiguration and provides means to configure the tonal range.
    """


@dataclass
class TonalRangeAuto(TonalRange):
    """Automatic tonal range configuration.

    Inherits from TonalRange and provides automatic adjustment of the tonal range.
    """


@dataclass
class TonalRangeManual(TonalRange):
    """Manual tonal range configuration.

    Inherits from TonalRange and provides manual adjustment of the tonal range.
    """


@dataclass
class Gamma(OptionalConfiguration):
    """Represents the gamma configuration for a camera.

    Inherits from OptionalConfiguration and provides means to configure the gamma settings.
    """


@dataclass
class GammaAuto(TonalRange):
    """Automatic gamma configuration.

    Inherits from Gamma and provides automatic adjustment of the gamma settings.
    """


@dataclass
class GammaManual(TonalRange):
    """Manual gamma configuration.

    Inherits from Gamma and provides manual adjustment of the gamma settings.

    Attributes:
        value (float): The gamma value to be set manually.
    """


class AutoMode(Enum):
    """
    Note: off can comprise manual (if implemented).
    """
    ONCE = 'Once'
    CONTINUOUS = 'Continuous'
    OFF = 'Off'


@dataclass
class Gain(OptionalConfiguration):
    """Holds means of setting the camera gain.

    Increasing the gain increases all pixel values of the image.

    Attributes:
        digital: Union[bool, NotAvailable]:
            Distinguishes between analog and digital gain. Analog gain is applied before the signal from the camera
            sensor is converted into digital values. Digital gain is applied after the conversion, i.e., it is
            basically a multiplication of the digitized values.
    """

    digital: Union[bool, NotAvailable]


@dataclass
class ManualGain(Gain):
    """Manual gain configuration.

    Attributes:
        value: Union[float, int]: The configured gain. On some camera models, the gain must be entered as a "raw"
                                  value on an integer scale. The camera needs the raw value for its internal processing
                                  mechanism. The raw value, however, isn't the same as the actual gain value, which
                                  is expressed in decibels (dB).
    """
    value: Union[float, int]


@dataclass
class AutoGain(Gain):
    """ Automatically adjust the gain until a target brightness value, i.e., an average gray value, has been reached
    on a region of interest .


    Attributes:
        upper_limit: Union[float, int, NotAvailable]: Upper limit for the gain. Raw integer value on a custom scale or
                                                      actual gain value (float) expressed in decibels.
        lower_limit: Union[float, int, NotAvailable]: Lower limit for the gain. Raw integer value on a custom scale or
                                                      actual gain value (float) expressed in decibels.

        temporary (bool): If true, apply the mode only temporarily.
    """
    upper_limit: Union[float, int, NotAvailable]
    lower_limit: Union[float, int, NotAvailable]
    temporary: bool = False


@dataclass
class Exposure(OptionalConfiguration):
    """Holds means of setting the exposure time."""


@dataclass
class AutoExposure(Exposure):
    """Automatically adjust the exposure time within specified limits until a target brightness has been reached.

    Attributes:
        upper_limit: (Union[int, NotAvailable]) Upper limit for the exposure time [microseconds].
        lower_limit: (Union[int, NotAvailable]) Lower limit for the exposure time [microseconds].

        temporary (bool): If true, apply the mode only temporarily.

    See:
        Brightness: If AutoGain is configured simultaneously, an auto function profile specifies how the both are
                    balanced.
    """

    upper_limit: Union[float, int, NotAvailable]
    lower_limit: Union[float, int, NotAvailable]
    temporary: bool = False


@dataclass
class TimedExposure(Exposure):
    """In this mode, the length of exposure is defined by the value of the camera's Exposure Time setting.

    Attributes:
        time (float) exposure time in microseconds.
    """
    time: float


@dataclass
class TriggeredExposure(Exposure):
    """In this mode, the length of exposure is defined by the width of the hardware trigger signal, or
    by a pair of software or hardware start and end triggers.

    Attributes:
        trigger_exposure_start: Union[TriggerMode] trigger that either indicates exposure start, or exposure start
                                                      and stop if the trigger has a duration.

        trigger_exposure_end: Union[TriggerMode] trigger that indicates stop of the exposure.

        start_delay: Union[int]:  Determines the period of time between the detection of the trigger signal and the
                                     actual start of the exposure.

        overlap_time_max: Union[float]: With overlapping image acquisition, exposure of the next image begins when
                                           the camera is reading out the sensor data of the previous image.
                                           In triggered exposure mode, the camera does not know how long the image is
                                           exposed. Overlap_time_max can be used to indicate the shortest exposure time
                                           that is intended to use be used in microseconds.

    """
    trigger_exposure_start: Union[TriggerMode, NotAvailable]
    trigger_exposure_end: Union[TriggerMode, NotAvailable]
    start_delay: Union[int, float, NotAvailable] = field(default_factory=NotAvailable)
    overlap_time_max: Union[float, NotAvailable] = field(default_factory=NotAvailable)


@dataclass
class BlackLevel(OptionalConfiguration):
    """Dark current can have a negative impact on image quality.

    Black level compensation / adjustment can mitigate this.
    """


@dataclass
class ManualBlackLevel(BlackLevel):
    """
    Adjusting the camera's black level will result in an offset to the pixel's gray values
    output by the camera.

    Attributes:
        value (float): manual shift of the black level.
    """
    value: float


@dataclass
class Brightness(OptionalConfiguration):
    """Holds the brightness configuration.

    Attributes:
        exposure (Exposure): Holds the current exposure time configuration.

        gain (Gain): Holds the current gain configuration.

        auto_roi_configuration: Union[str, Roi, NotAvailable]
            Roi which is used to assess gain configuration.
            If empty and the sensor supports roi configuration, use the entire image as roi.

        auto_target_brightness: Union[float, NotAvailable]:
            Brightness that is to be attained for automatic brightness control.

        auto_target_brightness_damping: Union[float, NotAvailable]:
            Specifies how quick the target brightness value is adjusted.
            Specifying a lower value leads to a quicker conversion to the target brightness at the cost
            of potential image instability and brightness oscillation.

        auto_function_profile: Union[str, NotAvailable]:
            If available, this specifies how gain and exposure time are balanced when
            automatic adjustment is configured for both of them.

        auto_backlight_compensation (Union[float, NotAvailable])
            If there is a bright light behind your image target, the target may be underexposed
            and appear silhouetted.
            The Backlight Compensation feature allows the camera to compensate for this effect.

            When the feature is enabled, a given percentage of the brightest pixels in the image
            (i.e., the pixels with the highest pixel values) are not taken into account for the
            target brightness value calculations performed by the "Gain Auto" and "Exposure Auto"
            auto functions.

            For example, setting the parameter value to 0.3 means that 30 % of the brightest
            pixels in the image will not be taken into account for the target value calculations.

            This allows the camera to properly expose the darker regions of the image.
    """

    exposure: Union[Exposure, NotAvailable] = field(default_factory=NotAvailable)
    gain: Union[Gain, NotAvailable] = field(default_factory=NotAvailable)
    black_level: Union[BlackLevel, NotAvailable] = field(default_factory=NotAvailable)

    auto_roi_configuration: Union[str, Roi, NotAvailable] = field(default_factory=NotAvailable)
    auto_target_brightness: Union[float, NotAvailable] = field(default_factory=NotAvailable)
    auto_target_brightness_damping: Union[float, NotAvailable] = field(default_factory=NotAvailable)
    auto_function_profile: Union[str, NotAvailable] = field(default_factory=NotAvailable)
    auto_backlight_compensation: Union[float, NotAvailable] = field(default_factory=NotAvailable)

    @property
    def auto_function_enabled(self) -> bool:
        return isinstance(self.gain, AutoGain) or isinstance(self.exposure, AutoExposure)


@dataclass
class WhiteBalance(OptionalConfiguration):
    """
    This class contains means of correcting color shifts so that white objects appear white in acquired images.

    References:
        [1] Balance White. (2024).
            Retrieved from https://docs.baslerweb.com/balance-white
    """
    ...


@dataclass
class AutoWhiteBalance(WhiteBalance):
    """Automatic white balance that is performed on a region of interest.

    Automatic white balancing can be performed in an iterative process on RGB data (h x w x c).
    1. Capture a new image
        >>> image: np.ndarray
        >>> balance_ratio: List[float] = [1.0, ] * 3

    2. Determine the color with the highest average value.
       >>> image.shape = (-1, 3)
       >>> argmax_color = np.argmax(np.mean(image, axis=0))

       Set the balance ratio for color `c` to 1.
       >>> balance_ratio[argmax_color] = 1.0

    3. Adjust the balance ratio for the other two colors until the average gray values for red, green, and blue are
       identical. Because the value with the highest average value is seleted in the previous step, the balance ratios
       for the two other colors will be >= 1.

       The adjustment_damping factor is considered when

    Attributes:
        roi_configuration Union[str, Roi, NotAvailable] The roi on which the auto white balance is executed
        adjustment_damping (Union[float, NotAvailable]):  If specified, colors are not adjusted immediately, but after
            a certain delay. This can be useful to avoid the automatic control mechanism of the auto functions being
            disrupted by objects moving in and out of the camera's area of view.

    References:
        [1] Balance White Auto. (2024).
            Retrieved from https://docs.baslerweb.com/balance-white-auto

    """

    roi_configuration: Union[str, Roi, NotAvailable]
    adjustment_damping: Union[float, NotAvailable]
    temporary: bool = False


@dataclass
class ManualWhiteBalance(WhiteBalance):
    """Digital gain correction that is applied per color channel.

    Attributes:
        channels (Dict[str, float]): Contains the balance ratio for each channel. Typically,
                the balance ratios are configured as follows:
                balance_ratio == 1: The intensity of the color remains unchanged.
                               > 1: The intensity of the color is increased relative to the other two colors.
                               < 1: The intensity of the color is decreased relative to the other two colors.
                However, some cameras don't allow balance ratios smaller than one, because a balance
                can be normalized to an equivalent balance by dividing through the sum of all
                balance ratios or multiplying each value with a specific factor.
    """
    channels: Dict[str, float]


@dataclass
class ColorAdjustmentConfiguration(OptionalConfiguration):
    """Color adjustment configuration. Can be used for specific colors or globally.

    Attributes:
        hue_rotation (Union[float, int, NotAvailable]):
            rotation of the RGB color cube, typically in degrees.

        saturation_adjustment: (Union[float, NotAvailable]):
            Lower parameter values result in more muted colors that are closer to gray.
            Higher parameter values result in more vivid, vibrant colors.

    """
    hue_rotation: Union[float, int, NotAvailable]
    saturation_adjustment: Union[float, NotAvailable]


@dataclass
class ColorAdjustment(OptionalConfiguration):
    """Holds the color adjustment configuration of the hue and saturation.

    Attributes:
        hue_rotation (Union[float, int, NotAvailable]):
            rotation of the RGB color cube, typically in degrees.

        saturation_adjustment: (Union[float, NotAvailable]):
            Lower parameter values result in more muted colors that are closer to gray.
            Higher parameter values result in more vivid, vibrant colors.

        color_adjustments (Dict[str, ColorAdjustmentConfiguration]):
            hue, saturation adjustment for each color.
    """
    hue_rotation: Union[float, int, NotAvailable]
    saturation_adjustment: Union[float, NotAvailable]
    color_adjustments: Dict[str, Union[ColorAdjustmentConfiguration, NotAvailable]]


@dataclass
class Binning(OptionalConfiguration):
    """Binning combines sensor pixel values into a single value.

    This may increase the signal-to-noise ratio, the camera's response to light, and its frame rate.

    Attributes:
        vertical (int) number of pixels to bin vertically.

        vertical_mode (Union[str, NotAvailable]): The operation that is performed to transform multiple
                                                  pixels into one vertically (e.g., sum).

        horizontal (int) number of pixels to bin horizontally.

        horizontal_mode (Union[str, NotAvailable]): The operation that is performed to transform multiple
                                                    pixels into one horizontally (e.g., sum).
    """
    vertical: int = 1
    vertical_mode: Union[str, NotAvailable] = None
    horizontal: int = 1
    horizontal_mode: Union[str, NotAvailable] = None


@dataclass
class ContrastAdjustments(OptionalConfiguration):
    """
    Adjusting the contrast changes the degree of difference between light and dark areas in the image. The
    more contrast you apply, the more pronounced the difference will be.

    How the contrast is applied depends on the contrast mode that you select.

    Attributes:
        contrast_mode (Union[str, NotAvailable]): The camera uses a function to adjust the contrast,
                                                  that maps the input values to output values to adjust the CRF.
                                                  If supported, the user might be able to chose between a
                                                  linear function and other function shapes

        brightness_pivot (float): Adjust the brightness pivot, often between -1 and 1 and defaulted to
                          zero, indicating that the brightness pivot remains unchanged.

        contrast_pivot (float): The more contrast you apply, the more pronounced the difference will be.

    """
    contrast_mode: Union[str, NotAvailable]
    brightness_pivot: float
    contrast_pivot: float


@dataclass
class CheckerPatternRemoval(metaclass=ABCMeta):
    """
    Some cameras produce images that display artifacts resembling a superimposed checker pattern due to the design
    of the sensor.
    Any 2 x 2 region of pixels in the sensor array respond identically to light only if its incidence is perpendicular
    to the sensor's surface; if the light arrives at a slightly oblique angle, they will respond slightly differently,
    thus causing a checker pattern

    When running pattern removal, ideally follow the following steps:
        1. set the image roi to its maximum size
        2. establish homogenous illumination
        3. Disable automatic modes (gain auto, exposure auto, sequencer).

    Should be repeated once one of the following factors change:
        1. Optical system (lens, aperture, focus)
        2. illumination
        3. features like exposure time, binning, roi, black level and digital shift
    """


@dataclass
class CheckerPatternRemovalAuto(CheckerPatternRemoval):
    """Automatic checker pattern removal.

    Inherits from CheckerPatternRemoval and provides automatic removal of checker patterns.
    """


@dataclass
class CheckerPatternRemovalManual(CheckerPatternRemoval):
    """Manual checker pattern removal.

    Inherits from CheckerPatternRemoval and provides manual removal of checker patterns.
    """


@dataclass
class Camera(Sensor):
    """Contains all available camera properties.

    Attributes:
        acquisition_mode (Union[AcquisitionMode, NotAvailable])
            Describes how to interact with the sensor for gathering images.
            Examples may include Triggered acquisition or continuous acquisition.
            In triggered acquisition, a trigger (software or hardware) is required for gathering an image.

        operation_mode (Union[OperationMode, NotAvailable]):
            Holds a sensor-specific operation, that changes the mode in which the camera
            operates independent of the pixel format and the acquisition mode.
            Examples may include short range operation mode in time of flight sensors.
            In this example, the acquisition mode, channels, the number of pixels and the pixel
            format stay the same, but the sensor gathers only information on pixels that are
            close to the sensor.

        pixel_format (Union[PixelFormat, NotAvailable]):
            Holds information on the acquired data, including the available channels, the bit depth, the color space
            properties and more.

        current_camera_configuration: Union[str, NotAvailabel]:
            Some cameras allow saving the current configuration on the camera.
            This value holds the name of the currently active configuration.
            Some configurations might be editable, others might be read-only.
            For specific vendors, the configuration is called a user set.

        startup_camera_configuration: Union[str, NotAvailable]
            The configuration (user set) that is automatically loaded when the camera starts.

        resulting_acquisition_rate (float):
            Estimated effective line  / frame rate with the current camera settings in Hz.

        crop (Crop):

            Allows you to specify the part of the sensor array that you want to use for image
            acquisition.

            If an image crop has been specified, the camera will only transmit pixel data from
            within that region.
            On most cameras, this increases the camera's maximum frame rate significantly.

        binning (Union[Binning, NotAvailable]):
            Holds the binning configuration, that allows combining multiple sensor pixel values into a single value.

        rois:
            Holds settings on what part of the sensor array specific camera functions are
            controlled. Examples can include auto white balance features.

            ROIs can be used for multiple purposes.

        blooming_reduction (Union[bool, NotAvailable]):
            Reduce image artifacts caused by blooming.
    """
    #
    # Operation variables
    #
    acquisition_mode: Union[AcquisitionMode, NotAvailable] = field(default_factory=NotRetrieved)
    operation_mode: Union[str, NotAvailable] = field(default_factory=NotRetrieved)
    pixel_format: Union[PixelFormat, NotAvailable] = field(default_factory=NotRetrieved)
    current_camera_configuration: Union[int, str, NotAvailable] = field(default_factory=NotRetrieved)
    startup_camera_configuration: Union[int, str, NotAvailable] = field(default_factory=NotRetrieved)

    #
    # Reporting (read-only variables, statistics)
    #
    resulting_acquisition_rate: Union[float, NotAvailable] = field(default_factory=NotRetrieved)

    #
    # Settings that have an impact on the resolution
    #
    crop: Union[Roi, NotAvailable] = field(default_factory=NotRetrieved)
    binning: Union[Binning, NotAvailable] = field(default_factory=NotRetrieved)

    #
    # Pixel level settings
    #
    brightness: Union[Brightness, NotAvailable] = field(default_factory=NotAvailable)
    white_balance: Union[WhiteBalance, NotAvailable] = field(default_factory=NotRetrieved)
    contrast_adjustments: Union[ContrastAdjustments, NotAvailable] = field(default_factory=NotRetrieved)
    color_adjustments: Union[ColorAdjustment, NotAvailable] = field(default_factory=NotRetrieved)
    gamma: Union[Gamma, NotAvailable] = field(default_factory=NotRetrieved)
    # Artifact removal
    blooming_reduction: Union[bool, NotAvailable] = field(default_factory=NotRetrieved)
    pattern_removal: Union[CheckerPatternRemoval, NotAvailable] = field(default_factory=NotRetrieved)
    tonal_range: Union[TonalRange, NotAvailable] = field(default_factory=NotRetrieved)

    #
    # Filter
    #
    # spatial_filter: Union[int, NotAvailable] = field(default_factory=NotRetrieved)
    # temporal_filter: Union[int, NotAvailable] = field(default_factory=NotRetrieved)

    # reverse_x: Union[bool, NotAvailable] = field(default_factory=NotRetrieved)
    # reverse_y: Union[bool, NotAvailable] = field(default_factory=NotRetrieved)

    # Compression

    # Auto rois: TODO: Fill these values.
    rois: Dict[str, Union[Roi, Optional[NotAvailable]]] = field(default_factory=dict)

    trigger_timestamp: Optional[int] = None
    acquisition_timestamp: Optional[int] = None
    trigger_count: Optional[int] = None

    def __str__(self):
        info_str: str = f'{Sensor.__str__(self)}: '
        if is_available(self.acquisition_mode):
            info_str += f'mode {self.acquisition_mode}'
        else:
            info_str += 'mode n/a'
        if is_available(self.operation_mode):
            info_str += f' {self.operation_mode}'
        if is_available(self.pixel_format):
            info_str += f' {self.pixel_format.designation()}'
        if is_available(self.startup_camera_configuration):
            info_str += f' {self.startup_camera_configuration}->'
        if is_available(self.current_camera_configuration):
            info_str += f'{self.current_camera_configuration}'
        return info_str

    def __post_init__(self) -> None:
        Sensor.__post_init__(self)
        self.camera_model: CameraModel = get_camera_model(self)

    def get_used_triggers(self) -> Dict[str, TriggerMode]:
        # TODO: The timer and counter sources need to be considered as well
        used_sources: Dict[str, TriggerMode] = {}
        if isinstance(self.acquisition_mode, TriggerAcquisitionMode):
            for trigger in filter(is_available, (self.acquisition_mode.trigger_start,
                                                 self.acquisition_mode.trigger_frame_end,
                                                 self.acquisition_mode.trigger_frame_burst_start,
                                                 self.acquisition_mode.trigger_frame_burst_end)):
                used_sources[trigger.trigger_source] = trigger

        if self.brightness is not None:
            if isinstance(self.brightness.exposure, TriggeredExposure):
                for trigger in filter(is_available,
                                      (self.brightness.exposure.trigger_exposure_start,
                                       self.brightness.exposure.trigger_exposure_end)):
                    used_sources[trigger.trigger_source] = trigger

        return used_sources

    def get_free_software_triggers(self) -> List[str]:
        used_triggers: Dict[str, TriggerMode] = self.get_used_triggers()
        return [source for source in self.camera_model.software_trigger_sources_generic if source not
                in used_triggers]

    def get_free_hardware_triggers(self) -> List[str]:
        used_triggers: Dict[str, TriggerMode] = self.get_used_triggers()
        return [source for source in self.camera_model.gpio_lines if source not in used_triggers]

    def is_hardware_trigger(self, source) -> bool:
        return source is not None and source in self.camera_model.gpio_lines

    def is_software_trigger(self, source) -> bool:
        return source is not None and source in self.camera_model.software_trigger_sources_generic

    def get_free_roi(self) -> Set[str]:
        """
        Returns the next region of interest from the list of configurable regions of interest,
        that is unused.
        """
        available_rois = set(self.camera_model.rois)

        if isinstance(self.white_balance, AutoWhiteBalance):
            try:
                available_rois.remove(self.white_balance.roi_configuration)
            except KeyError:
                pass

        if self.brightness is not None:
            if self.brightness.auto_function_enabled:
                available_rois.remove(self.brightness.auto_roi_configuration)

        return available_rois


_camera_models: Dict[SensorModelDescriptor, CameraModel] = {SensorModelDescriptor(None): CameraModel()}


def register_camera_model(model_descriptor: SensorModelDescriptor, model: CameraModel) -> None:
    # Check for conflicting information during registration

    if model_descriptor in _camera_models:
        raise ValueError(f'{model_descriptor} is already registered.')

    _camera_models[model_descriptor] = model


def get_camera_model(camera: Camera) -> CameraModel:
    if not len(_camera_models):
        raise ValueError

    max_score = None
    max_model = None
    for model_descriptor, model in _camera_models.items():
        score = model_descriptor.score(camera.identifier, camera.model, camera.manufacturer)
        if max_score is None or score > max_score:
            max_score, max_model = score, model

    return copy.deepcopy(max_model)


# TODO: Camera model management! How to manage? Where to store?

# https://www2.baslerweb.com/media/downloads/documents/brochure/BAS2304_Features_Check_List_web.pdf
basler_camera_model = CameraModel(
    available_pixel_formats=[GenicamPixelFormat.from_name('Mono8'), GenicamPixelFormat.from_name('Mono12'),
                             GenicamPixelFormat.from_name('Mono12p')],
    available_configuration_names_read_only=['Default', 'HighGain', 'AutoFunctions'],
    available_configuration_names_writable=['UserSet1', 'UserSet2', 'UserSet3'],
    rois=['ROI1', 'ROI2'],
    auto_function_profiles=['MinimizeGain', 'MinimizeExposureTime', 'MinimizeGainQuick',
                            'MinimizeExposureTimeQuick', 'Smart', 'AntiFlicker50Hz',
                            'AntiFlicker60Hz'],

    gpio_lines=['Line1', 'Line2', 'Line3', 'Line4'],
    software_trigger_sources_generic=['SoftwareSignal1', 'SoftwareSignal2', 'SoftwareSignal3'],
    software_trigger_frame_start='Software',

    binning_modes=['Sum', 'Average'],
    contrast_modes=['Linear', 'SCurve'],

    frame_start_trigger='FrameStart',
    frame_stop_trigger=None,
    frame_active_trigger=None,

    frame_burst_start_trigger='FrameBurstStart',
    frame_burst_stop_trigger=None,
    frame_burst_active_trigger=None,

    exposure_start_trigger=None,
    exposure_stop_trigger=None,
    exposure_active_trigger=None,

    line_start_trigger=None,
)

register_camera_model(SensorModelDescriptor('Basler', 'acA4112'), basler_camera_model)

# KEYENCE CORPORATION, CA-H500MX,


class SensorStateException(SensorException):

    @staticmethod
    def from_sensor(sensor: Sensor, expected_states: Optional[Set[SensorState]] = None):
        return SensorException(f'Sensor {sensor.identifier} is in invalid state {sensor.state}.' +
                               ('' if not expected_states else
                                f'Allowed states: {", ".join(map(str, expected_states))}.'))


class SensorStateInconsistent(SensorException):
    ...


@dataclass
class CameraData(SensorData):
    content: List[np.ndarray]

    gain: Optional[float] = None
    exposure_time: Optional[Union[int, float]] = None
    timestamp: Optional[int] = None

    meta_information: Dict[str, Any] = field(default_factory=dict)
