import numpy as np
from abc import ABCMeta, abstractmethod
from enum import Enum
from dataclasses import dataclass
from typing import Union, List, Tuple, Dict, Optional, get_args
from logging import getLogger

from dataspree.inference.sensor.data.pixel_format.utils import join_designators, all_implementations
from dataspree.inference.sensor.data.pixel_format.number_of_bits import Size

logger = getLogger(__name__)


class CompoundChannel(Enum):
    RAW = 'Raw'
    MONO = 'Mono'
    CONFIDENCE = 'Confidence'
    RED = 'R'
    GREEN = 'G'
    BLUE = 'B'
    LUMA = 'Y'
    ALPHA = 'a'
    ALPHA_2 = 'A'
    HUE = 'H'
    SATURATION = 'S'
    VALUE = 'V'  # or color difference for YUV
    CHROMA = 'U'
    WHITE = 'W'
    INTENSITY = 'I'
    BLUE_DIFFERENCE = 'Cb'
    RED_DIFFERENCE = 'Cr'
    CYAN = 'C'
    MAGENTA = 'M'
    YELLOW = 'Ye'
    INFRARED = 'Ir'


single_channel_priority_mapping: Dict[CompoundChannel, float] = {
    CompoundChannel.INTENSITY: 1,
    CompoundChannel.LUMA: 1,
    CompoundChannel.MONO: 1,
    CompoundChannel.VALUE: 2,
    CompoundChannel.GREEN: 2,
    CompoundChannel.RED: 3,
    CompoundChannel.BLUE: 3,
    CompoundChannel.SATURATION: 2,
}


class CoordinateChannel(Enum):
    COORD3D_A = 'A'
    COORD3D_B = 'B'
    COORD3D_C = 'C'


_custom_channels: List[str] = []

Channel = Union[CompoundChannel, CoordinateChannel, str]

_reverse_channel_mapping: Dict[str, Channel] = {f.value: f for f in CompoundChannel}

if len(_reverse_channel_mapping) != len(CompoundChannel):
    raise ValueError('Invalid channel mapping.')


def is_channel(potential_channel):
    for arg in get_args(Channel):
        if isinstance(potential_channel, arg):
            return True

    return False


def channel_name(channel: Channel) -> str:
    if isinstance(channel, str):
        return channel
    if isinstance(channel, CompoundChannel):
        return channel.value
    return channel.value


@dataclass
class ChannelProperty:
    combinable: bool


channel_properties: Dict[Channel, ChannelProperty] = {
    CompoundChannel.RAW: ChannelProperty(False),
    CompoundChannel.MONO: ChannelProperty(False),
    CompoundChannel.CONFIDENCE: ChannelProperty(False),
    CoordinateChannel.COORD3D_A: ChannelProperty(True),
    CoordinateChannel.COORD3D_B: ChannelProperty(True),
    CoordinateChannel.COORD3D_C: ChannelProperty(True),
    CompoundChannel.RED: ChannelProperty(True),
    CompoundChannel.GREEN: ChannelProperty(True),
    CompoundChannel.BLUE: ChannelProperty(True),
    CompoundChannel.LUMA: ChannelProperty(True),
    CompoundChannel.ALPHA: ChannelProperty(True),
    CompoundChannel.ALPHA_2: ChannelProperty(True),
    CompoundChannel.HUE: ChannelProperty(True),
    CompoundChannel.SATURATION: ChannelProperty(True),
    CompoundChannel.VALUE: ChannelProperty(True),
    CompoundChannel.CHROMA: ChannelProperty(True),
    CompoundChannel.WHITE: ChannelProperty(True),
    CompoundChannel.INTENSITY: ChannelProperty(True),
    CompoundChannel.BLUE_DIFFERENCE: ChannelProperty(True),
    CompoundChannel.RED_DIFFERENCE: ChannelProperty(True),
    CompoundChannel.CYAN: ChannelProperty(True),
    CompoundChannel.MAGENTA: ChannelProperty(True),
    CompoundChannel.YELLOW: ChannelProperty(True),
    CompoundChannel.INFRARED: ChannelProperty(True),
}

if len(channel_properties) != len(CompoundChannel) + len(CoordinateChannel):
    raise ValueError


def register_custom_channel(name: str, channel_property: ChannelProperty):
    if name in _custom_channels:
        raise ValueError

    if name in channel_property:
        raise ValueError

    if name in _reverse_channel_mapping:
        raise ValueError

    _custom_channels.append(name)
    channel_properties[name] = channel_property
    _reverse_channel_mapping[name] = name


def get_channels() -> List[Channel]:
    return _custom_channels + list(CompoundChannel)


class Channels(metaclass=ABCMeta):
    @abstractmethod
    def name(self, indices: Optional[List[int]] = None) -> str:
        ...

    @abstractmethod
    def __len__(self) -> int:
        ...

    @abstractmethod
    def __getitem__(self, item) -> Channel:
        ...

    @staticmethod
    def create(channel_start: str, allow_repeat: bool = False) \
            -> List[Tuple['Channels', Optional['ColorSpaceStandard'], str]]:
        possible_channels: List[Tuple[Channels, str]] = []
        for implementation in all_implementations(Channels):
            try:
                possible_channels.extend(implementation.from_designation(channel_start, allow_repeat))
            except Exception as e:
                logger.warning(f'An exception occurred: {e} {implementation.__name__}')

        compos: List[Tuple[Channels, Optional[ColorSpaceStandard], str]] = []
        for channels, substring in possible_channels:
            optional_standards = ColorSpaceStandard.create(substring, channels)
            for cs, remaining_substring in optional_standards:
                compos.append((channels, cs, remaining_substring))
        return compos

    @classmethod
    @abstractmethod
    def from_designation(cls, channel_start: str, allow_repeat: bool = False) -> List[Tuple['Channels', str]]:
        ...


class CompoundChannels(Channels):
    def __init__(self, channels: List[Channel], allow_repeat: bool = False) -> None:
        self.channels: List[Channel] = channels

        for channel in self.channels:
            if channel not in channel_properties:
                raise ValueError

            channel_property = channel_properties[channel]
            if not channel_property.combinable and len(self.channels) > 1:
                raise ValueError

        if len(set(self.channels)) != len(self.channels) and not allow_repeat:
            raise ValueError('Repeat channel.')

    def __len__(self) -> int:
        return len(self.channels)

    def __getitem__(self, item) -> Channel:
        return self.channels[item]

    def name(self, indices: Optional[List[int]] = None) -> str:
        indices = list(range(len(self.channels))) if indices is None else indices
        return ''.join(map(channel_name, [self.channels[i] for i in indices]))

    @classmethod
    def from_designation(cls, channel_start: str, allow_repeat: bool = False) -> List[Tuple['CompoundChannels', str]]:
        if not channel_start:
            return []

        if channel_start[0] == '_':
            channel_start = channel_start[1:]

        channels: List[Tuple[CompoundChannels, str]] = []

        for channel_identifier, channel in _reverse_channel_mapping.items():
            if channel_start.startswith(channel_identifier):
                new_start = channel_start[len(channel_identifier):]
                channels.append((CompoundChannels([channel], allow_repeat=allow_repeat), new_start))
                for lc, start in cls.from_designation(new_start, allow_repeat):
                    channels.append((CompoundChannels([channel, *lc.channels], allow_repeat=allow_repeat), start))

        return channels

    def __eq__(self, other):
        return isinstance(other, CompoundChannels) and self.channels == other.channels


class Coord3d(CompoundChannels):

    def __init__(self, channels: List[Channel], allow_repeat: bool = False) -> None:
        CompoundChannels.__init__(self, channels, allow_repeat=allow_repeat)
        for channel in channels:
            if channel not in (CoordinateChannel.COORD3D_A, CoordinateChannel.COORD3D_B,
                               CoordinateChannel.COORD3D_C):
                raise ValueError

    def name(self, indices: Optional[List[int]] = None):
        return f'Coord3D_{CompoundChannels.name(self, indices)}'

    @classmethod
    def from_designation(cls, channel_start, allow_repeat: bool = False) -> List[Tuple[Channels, str]]:
        if not channel_start:
            return []

        if channel_start[0] == '_':
            channel_start = channel_start[1:]

        if not channel_start.startswith('Coord3D_'):
            return []

        channel_start = channel_start[8:]
        channels = []
        for c in channel_start:
            if c == 'A':
                channels.append(CoordinateChannel.COORD3D_A)
            elif c == 'B':
                channels.append(CoordinateChannel.COORD3D_B)
            elif c == 'C':
                channels.append(CoordinateChannel.COORD3D_C)
            else:
                break

        if not len(channels):
            return []

        return [(Coord3d(channels, allow_repeat=allow_repeat), channel_start[len(channels):])]


class ColorSpaceStandard(metaclass=ABCMeta):
    def __init__(self, name: str) -> None:
        self.name = name

    @classmethod
    def initialize(cls, channels: Channels) -> 'ColorSpaceStandard':
        ...

    @classmethod
    def from_designation(cls, designation_substring: str, channels: Channels) -> List[Tuple['ColorSpaceStandard', str]]:
        if not designation_substring:
            return []

        if designation_substring[0] == '_':
            designation_substring = designation_substring[1:]

        remaining_substring = cls.compatible_with(channels, designation_substring)
        if remaining_substring is not None:
            return [(cls.initialize(channels), remaining_substring)]

        return []

    @staticmethod
    @abstractmethod
    def compatible_with(channels: Channels, designation_substring: Optional[str] = None) -> Optional[str]:
        ...

    @staticmethod
    def create(designation_substring: str, channels: Channels) -> List[Tuple[Optional['ColorSpaceStandard'], str]]:
        compos: List[Tuple[Optional[ColorSpaceStandard], str]] = [(None, designation_substring)]
        for standard_implementation in all_implementations(ColorSpaceStandard):
            try:
                compos.extend(standard_implementation.from_designation(designation_substring, channels))
            except Exception as e:
                logger.warning(f'An exception occurred: {e} {standard_implementation.__name__}')
        return compos


class ItuRBt601(ColorSpaceStandard):

    def __init__(self, ) -> None:
        ColorSpaceStandard.__init__(self, '601')

    @classmethod
    def initialize(cls, channels: Channels) -> 'ItuRBt601':
        return cls()

    @staticmethod
    def compatible_with(channels: Channels, designation_substring: Optional[str] = None) -> Optional[str]:
        ident = '601'
        if (designation_substring is None or designation_substring.startswith(ident)) and channels.name() == 'YCrCb':
            return designation_substring[len(ident):]

    def __eq__(self, other):
        return isinstance(other, type(self))


class ItuRBt709(ColorSpaceStandard):

    def __init__(self) -> None:
        ColorSpaceStandard.__init__(self, '709')

    @classmethod
    def initialize(cls, channels: Channels) -> 'ItuRBt709':
        return cls()

    @staticmethod
    def compatible_with(channels: Channels, designation_substring: Optional[str] = None) -> Optional[str]:
        ident = '709'
        if (designation_substring is None or designation_substring.startswith(ident)) and channels.name() == 'YCrCb':
            return designation_substring[len(ident):]

    def __eq__(self, other):
        return isinstance(other, type(self))


class Bt601(ColorSpaceStandard):

    def __init__(self) -> None:
        ColorSpaceStandard.__init__(self, '601')

    @classmethod
    def initialize(cls, channels: Channels) -> 'Bt601':
        return cls()

    @staticmethod
    def compatible_with(channels: Channels, designation_substring: Optional[str] = None) -> Optional[str]:
        ident = '601'
        if (designation_substring is None or designation_substring.startswith(ident)) and channels.name() == 'YCbCr':
            return designation_substring[len(ident):]

    def __eq__(self, other):
        return isinstance(other, type(self))


class Bt709(ColorSpaceStandard):

    def __init__(self) -> None:
        ColorSpaceStandard.__init__(self, '709')

    @classmethod
    def initialize(cls, channels: Channels) -> 'Bt709':
        return cls()

    @staticmethod
    def compatible_with(channels: Channels, designation_substring: Optional[str] = None) -> Optional[str]:
        ident = '709'
        if (designation_substring is None or designation_substring.startswith(ident)) and channels.name() == 'YCbCr':
            return designation_substring[len(ident):]

    def __eq__(self, other):
        return isinstance(other, type(self))


@dataclass
class Location(metaclass=ABCMeta):
    @abstractmethod
    def location_designation(self, channels: Channels, color_space_name: str) -> str:
        ...

    @abstractmethod
    def location_name(self, channels: Channels) -> str:
        """
        Returns Positioning in Image

        """
        ...

    @abstractmethod
    def number_of_channels(self) -> int:
        ...

    @staticmethod
    def create(full_designation: str) -> List[Tuple['Location', Channels, ColorSpaceStandard, str]]:
        locs: List[Tuple[Location, Channels, ColorSpaceStandard, str]] = []
        for class_name in all_implementations(Location):
            try:
                locs.extend(class_name.from_designation(full_designation))
            except Exception as e:
                logger.warning(f'An exception occurred: {e} {class_name}')
        return locs

    @staticmethod
    @abstractmethod
    def from_designation(full_designation: str) -> List[Tuple['Location', Channels, ColorSpaceStandard, str]]:
        ...


@dataclass
class MonoLocation(Location):
    """
    If a monochrome camera uses one of the mono pixel formats, it outputs 8, 10, 12, or 14 bits of data per pixel.

    If a Basler color camera uses one of the mono pixel formats, the values for each pixel are first converted to the
    YCbCr color model and returns the Y component (brightness).
    This is equivalent to the value that would be derived from a pixel in a monochrome sensor.
    """

    def location_designation(self, channels: Channels, color_space_name: str) -> str:
        return join_designators(channels.name(), color_space_name)

    def location_name(self, channels: Channels) -> str:
        return 'Mono'

    def number_of_channels(self) -> int:
        return 1

    @staticmethod
    def from_designation(full_designation: str) -> List[Tuple['Location', Channels, Optional[ColorSpaceStandard], str]]:
        if not full_designation:
            return []
        if full_designation[0] == '_':
            full_designation = full_designation[1:]

        potential_channels: List[Tuple[Channels, Optional[ColorSpaceStandard], str]] = Channels.create(full_designation)
        mono_locations: List[Tuple[Location, Channels, Optional[ColorSpaceStandard], str]] = []
        for channels, standard, designation_suffix in potential_channels:
            if len(channels) == 1:
                mono_locations.append((MonoLocation(), channels, standard, designation_suffix))
        return mono_locations

    def __eq__(self, other):
        return isinstance(other, MonoLocation)


@dataclass
class LMNLocation(Location):
    fields: Tuple[int, ...]

    def __post_init__(self):
        if len(self.fields) == 0:
            raise ValueError

        if len(self.fields) >= len('LMNOPQRSTUVWXYZ'):
            raise ValueError

        if self.fields[0] != 4:
            raise ValueError

        for field in self.fields[1:]:
            if not (0 < field <= self.fields[0]):
                raise ValueError

    def location_designation(self, channels: Channels, color_space_name: str) -> str:
        channels = join_designators(channels.name(), color_space_name)
        if set(self.fields) == {4}:
            return channels

        return join_designators(channels, ''.join(map(str, self.fields)))

    def location_name(self, channels: Channels) -> str:
        return 'LMNOPQRSTUVWXYZ'[:len(self.fields)] + ''.join(map(str, self.fields))

    def number_of_channels(self) -> int:
        return len(self.fields)

    @staticmethod
    def from_designation(full_designation: str) -> List[Tuple['Location', Channels, Optional[ColorSpaceStandard], str]]:
        if not full_designation:
            return []
        if full_designation[0] == '_':
            full_designation = full_designation[1:]
        potential_channels: List[Tuple[Channels, Optional[ColorSpaceStandard], str]] = Channels.create(full_designation)

        lmn_locations: List[Tuple[Location, Channels, Optional[ColorSpaceStandard], str]] = []
        for channels, standard, designation_suffix in potential_channels:
            number_of_channels = len(channels)
            if len(channels) <= 1:
                continue

            if designation_suffix.startswith('_'):
                designation_suffix = designation_suffix[1:]

            lmn: List[int] = []
            if len(designation_suffix) >= number_of_channels and designation_suffix[0] == '4':
                for i in range(number_of_channels):
                    if designation_suffix[i] in ('1', '2', '4'):
                        lmn.append(int(designation_suffix[i]))
                    else:
                        break

            if len(lmn) == number_of_channels:
                lmn_locations.append(
                    (LMNLocation(tuple(lmn)), channels, standard, designation_suffix[number_of_channels:]))
            else:
                lmn_locations.append((LMNLocation((4,) * number_of_channels), channels, standard, designation_suffix))

        return lmn_locations

    def __eq__(self, other):
        return isinstance(other, LMNLocation) and self.fields == other.fields


@dataclass
class BayerLocation(Location):
    """
    Modern color camera sensors are equipped with a Bayer color filter.

    They record red, green, blue pixels in distinct areas and  interpolate in between.
    Interpolation can be proprietary.

    If instead of the interpolated n x m values, you would like to directly receive the bayer
    pattern, use a bayer pixel format. Each pixel is filtered to record only one of the colors.

     array and can output color
    images based on Bayer pixel formats.


    If a color camera uses one of the Bayer pixel formats, it outputs 8, 10, or 12 bits of data per pixel.
    Each pixel is filtered to record only one of the colors red, green, and blue. The pixel data is not processed or
    interpolated in any way. This type of pixel data is sometimes referred to as "raw" output.

    Which Bayer formats are available depends on your camera model.

    If a monochrome camera uses one of the mono pixel formats, it outputs 8, 10, 12, or 14 bits of data per pixel.

    If a Basler color camera uses one of the mono pixel formats, the values for each pixel are first converted to the
    YCbCr color model and returns the Y component (brightness).
    This is equivalent to the value that would be derived from a pixel in a monochrome sensor.
    """

    fields: Tuple[str, str, str, str]

    def __post_init__(self):
        if ''.join(sorted(list(set(self.fields)))) != 'LMN':
            raise ValueError

        self.indices: Tuple[int, int, int, int] = (
            ord(self.fields[0]) - 76, ord(self.fields[1]) - 76,
            ord(self.fields[2]) - 76, ord(self.fields[3]) - 76)

    def location_designation(self, channels: Channels, color_space_name: str) -> str:
        return join_designators(f'Bayer{channels.name(list(self.indices[:2]))}', color_space_name)

    def location_name(self, channels: Channels) -> str:
        return f'Bayer_{"".join(self.fields)}'

    def number_of_channels(self) -> int:
        return 3

    @staticmethod
    def from_designation(full_designation: str) -> List[Tuple['Location', Channels, Optional[ColorSpaceStandard], str]]:
        if not full_designation:
            return []
        if full_designation[0] == '_':
            full_designation = full_designation[1:]

        if not full_designation.startswith('Bayer'):
            return []

        full_designation = full_designation[5:]
        potential_channels: List[Tuple[Channels, Optional[ColorSpaceStandard], str]] = Channels.create(full_designation)

        bayer_locations: List[Tuple[Location, Channels, Optional[ColorSpaceStandard], str]] = []
        for channels, standard, designation_suffix in potential_channels:
            if len(channels) != 2:
                continue

            # Ground truth; we know that the order is always RGB, and that the Green channel is always the one that
            # is repeated for the bayer location.
            missing_components: List[Union[CompoundChannel.RED, CompoundChannel.GREEN, CompoundChannel.BLUE]] = []
            if CompoundChannel.RED not in channels:
                missing_components.append(CompoundChannel.RED)
            if CompoundChannel.GREEN not in channels:
                missing_components.append(CompoundChannel.GREEN)
            if CompoundChannel.BLUE not in channels:
                missing_components.append(CompoundChannel.BLUE)

            if len(missing_components) != 1 or missing_components[0] is CompoundChannel.GREEN:
                continue

            component_order = {CompoundChannel.RED: 'L', CompoundChannel.GREEN: 'M', CompoundChannel.BLUE: 'N'}
            channel_order: List[str] = [component_order[channels[i]] for i in range(len(channels))]
            missing_component_order: str = component_order[missing_components[0]]

            full_order: List[str] = channel_order + channel_order[::-1]
            if full_order[-1] != 'M':
                full_order[-1] = missing_component_order
            else:
                full_order[-2] = missing_component_order

            assert sorted(set(full_order)) == ['L', 'M', 'N']

            bayer_locations.append((BayerLocation((full_order[0], full_order[1], full_order[2], full_order[3])),
                                    CompoundChannels(
                                        [CompoundChannel.RED, CompoundChannel.GREEN, CompoundChannel.BLUE]),
                                    standard,
                                    designation_suffix))

        return bayer_locations

    def __eq__(self, other):
        return isinstance(other, BayerLocation) and self.fields == other.fields


@dataclass
class BiColor(Location):
    fields: Tuple[str, str, str, str]

    def __post_init__(self):
        self.indices: Tuple[int, int, int, int] = (
            ord(self.fields[0]) - 76, ord(self.fields[1]) - 76,
            ord(self.fields[2]) - 76, ord(self.fields[3]) - 76)

    def location_designation(self, channels: Channels, color_space_name: str) -> str:
        return join_designators(f'BiColor{channels.name(list(self.indices))}',
                                color_space_name)

    def location_name(self, channels: Channels) -> str:
        return f'BiColor_{"".join(self.fields)}'

    def number_of_channels(self) -> int:
        return len(set(self.fields))

    @staticmethod
    def from_designation(full_designation: str) -> List[Tuple['Location', Channels, Optional[ColorSpaceStandard], str]]:
        if not full_designation:
            return []
        if full_designation[0] == '_':
            full_designation = full_designation[1:]
        if not full_designation.startswith('BiColor'):
            return []

        full_designation = full_designation[7:]
        potential_channels: List[Tuple[Channels, Optional[ColorSpaceStandard], str]] = Channels.create(full_designation,
                                                                                                       True)

        bi_color_locations: List[Tuple[Location, Channels, Optional[ColorSpaceStandard], str]] = []
        for channels, standard, designation_suffix in potential_channels:
            if len(channels) != 4:
                continue

            component_order = {CompoundChannel.RED: 'L', CompoundChannel.GREEN: 'M', CompoundChannel.BLUE: 'N'}
            try:
                target: List[str] = [component_order[channels[i]] for i in range(len(channels))]
            except KeyError:
                continue

            bi_color_locations.append((
                BiColor((target[0], target[1], target[2], target[3])),
                CompoundChannels([CompoundChannel.RED, CompoundChannel.GREEN, CompoundChannel.BLUE]),
                standard, designation_suffix))

        return bi_color_locations

    def __eq__(self, other):
        return isinstance(other, BiColor) and self.fields == other.fields


@dataclass
class ColorFilterArray(Location, metaclass=ABCMeta):
    ...


@dataclass
class SquarePattern(ColorFilterArray, metaclass=ABCMeta):
    def __post_init__(self) -> None:
        raise NotImplementedError


@dataclass
class SparseColorFilterLocation(ColorFilterArray, metaclass=ABCMeta):
    """
    Sparse Color Filter is a color filter array that includes panchromatic pixels with the red,
    green and blue color components.
    Different tile patterns can be created.
    """
    fields: Tuple[str, str, str, str]

    def __post_init__(self):
        if self.fields[0] != self.fields[2]:
            raise ValueError

        if self.fields[1] == self.fields[3]:
            raise ValueError

        raise NotImplementedError


class Components(metaclass=ABCMeta):

    @abstractmethod
    def full_designation(self) -> str:
        ...

    @abstractmethod
    def channel_designation(self) -> str:
        ...

    @abstractmethod
    def location_name(self) -> str:
        ...

    @property
    def number_of_channels(self) -> int:
        return len(self)

    @abstractmethod
    def __len__(self):
        ...

    @abstractmethod
    def __getitem__(self, item) -> Channel:
        ...

    @staticmethod
    def create(full_designation: str) -> List[Tuple['Components', str]]:
        compos: List[Tuple[Components, str]] = []
        for implementation in all_implementations(Components):
            try:
                compos.extend(implementation.from_designation(full_designation))
            except Exception as e:
                logger.warning(f'An exception occurred: {e} {implementation}')
        return compos

    @staticmethod
    @abstractmethod
    def from_designation(full_designation: str) -> List[Tuple['Components', str]]:
        ...

    @abstractmethod
    def get_structured_array(self, channels: List[np.ndarray]) -> Optional[np.ndarray]:
        ...


class GenicamComponents(Components):

    def __init__(self, channels: Union[Channels, Channel, List[Channel]],
                 location: Optional[Location] = None,
                 color_space_standard: Optional[ColorSpaceStandard] = None) -> None:

        self.channels: Channels
        # @dev: isinstance(channels, Channel) not supported in Python3.8
        if is_channel(channels):
            self.channels = CompoundChannels([channels])
        elif isinstance(channels, list):
            self.channels = CompoundChannels(channels)
        else:
            self.channels = channels
        if not isinstance(self.channels, Channels):
            raise ValueError('Wrong argument')

        self.location: Location
        if location is not None:
            self.location = location
        else:
            if len(self.channels) == 1:
                self.location = MonoLocation()
            else:
                self.location = LMNLocation((4,) * len(channels))

        if self.location.number_of_channels() != len(self.channels):
            raise ValueError('Incompatible location')

        self.color_space_standard: Optional[ColorSpaceStandard] = color_space_standard

    def channel_designation(self) -> str:
        return self.channels.name()

    def full_designation(self) -> str:
        color_space_name: str = self.color_space_standard.name if self.color_space_standard is not None else ''
        return self.location.location_designation(self.channels, color_space_name)

    def location_name(self) -> str:
        return self.location.location_name(self.channels)

    def __len__(self) -> int:
        return len(self.channels)

    def __getitem__(self, item) -> Channel:
        return self.channels[item]

    def __eq__(self, other):
        if isinstance(other, GenicamComponents):
            return (self.channels == other.channels and self.location == other.location and
                    self.color_space_standard == other.color_space_standard)
        return False

    @staticmethod
    def from_designation(full_designation: str) -> List[Tuple[Components, str]]:
        if not full_designation:
            return []
        if full_designation[0] == '_':
            full_designation = full_designation[1:]
        compos: List[Tuple[Components, str]] = []
        for location_implementation in all_implementations(Location):
            try:
                channels_location = location_implementation.from_designation(full_designation)
                for location, channels, standard, substring in channels_location:
                    compos.append((GenicamComponents(channels, location, standard), substring))

            except Exception as e:
                logger.warning(f'An exception occurred: {e} {location_implementation.__name__}')
        return compos

    def get_structured_array(self, image_channels: List[np.ndarray]) -> Optional[np.ndarray]:
        dtypes: List[Tuple] = []

        if len(image_channels) != len(self.channels):
            return None

        for channel, image_channel in zip(self.channels, image_channels):
            name: str = channel_name(channel)
            dtypes.append((name, image_channel.dtype))

            if image_channel.shape != image_channels[0].shape:
                return None

        image_structured = np.empty(image_channels[0].shape, dtype=dtypes)
        for (key, _type), image in zip(dtypes, image_channels):
            image_structured[key] = image

        return image_structured


class CustomComponents(Components, metaclass=ABCMeta):
    ...


def bayer_rgb(location: Tuple[str, str, str, str],
              color_space_standard: Optional[ColorSpaceStandard] = None) -> GenicamComponents:
    return GenicamComponents([CompoundChannel.RED, CompoundChannel.GREEN, CompoundChannel.BLUE],
                             BayerLocation(location), color_space_standard=color_space_standard)


def bi_color_rgb(location: Tuple[str, str, str, str],
                 color_space_standard: Optional[ColorSpaceStandard] = None) -> GenicamComponents:
    return GenicamComponents([CompoundChannel.RED, CompoundChannel.GREEN, CompoundChannel.BLUE],
                             BiColor(location), color_space_standard=color_space_standard)


class SizedColorSpaceStandard(metaclass=ABCMeta):
    def __init__(self, name: str) -> None:
        self.name = name

    @classmethod
    @abstractmethod
    def from_designation(cls, designation_substring: str, comps: Components, sizes: Size) \
            -> List[Tuple['SizedColorSpaceStandard', str]]:
        ...

    @staticmethod
    @abstractmethod
    def compatible_with(comps: Components, sizes: Size) -> bool:
        ...

    @staticmethod
    def create(designation_substring: str, comps: Components, sizes: Size) \
            -> List[Tuple[Optional['SizedColorSpaceStandard'], str]]:

        compos: List[Tuple[Optional[SizedColorSpaceStandard], str]] = [(None, designation_substring)]
        for standard_implementation in all_implementations(SizedColorSpaceStandard):
            try:
                compos.extend(standard_implementation.from_designation(designation_substring, comps, sizes))
            except Exception as e:
                logger.warning(f'An exception occurred: {e} {standard_implementation.__name__}')
        return compos


class RGB10V1(SizedColorSpaceStandard):

    def __init__(self) -> None:
        SizedColorSpaceStandard.__init__(self, 'V1')

    @staticmethod
    def compatible_with(comps: Components, sizes: Size) -> bool:
        return comps.channel_designation() == 'RGB' and sizes.designation() == '10'

    @classmethod
    def from_designation(cls, designation_substring: str, comps: Components, sizes: Size) \
            -> List[Tuple[SizedColorSpaceStandard, str]]:
        if not designation_substring:
            return []

        if not cls.compatible_with(comps, sizes):
            return []

        if designation_substring.startswith('V1'):
            return [(cls(), designation_substring[len('V1'):])]
        return []

    def __eq__(self, other):
        return isinstance(other, RGB10V1)


class RGB12V1(SizedColorSpaceStandard):

    def __init__(self) -> None:
        SizedColorSpaceStandard.__init__(self, 'V1')

    @staticmethod
    def compatible_with(comps: Components, sizes: Size) -> bool:
        return comps.channel_designation() == 'RGB' and sizes.designation() == '12'

    @classmethod
    def from_designation(cls, designation_substring: str, comps: Components, sizes: Size) \
            -> List[Tuple[SizedColorSpaceStandard, str]]:
        if not designation_substring:
            return []

        if not cls.compatible_with(comps, sizes):
            return []

        if designation_substring.startswith('V1'):
            return [(cls(), designation_substring[len('V1'):])]
        return []

    def __eq__(self, other):
        return isinstance(other, RGB12V1)
