from enum import Enum
from typing import Dict, Optional, List, Tuple, Union

from dataspree.inference.sensor.data.pixel_format.utils import join_designators


class PaddingType(Enum):
    PACKED = 'p'
    GROUPED = 'g'
    UNPACKED = ''


_padding_type_names: Dict[str, PaddingType] = {f.value: f for f in PaddingType}


def register_padding_type_name(name: str, padding_type: PaddingType) -> None:
    if name in _padding_type_names:
        raise ValueError

    _padding_type_names[name] = padding_type


# GGigE Vision 2.0 calls this Mono10Packed.
# We add this, even though this does not adhere to the PFNC.
register_padding_type_name('Packed', PaddingType.PACKED)


class Packing:
    def __init__(self, padding_type: Optional[Union[PaddingType, str]] = None, lsb: bool = True,
                 alignment: Optional[int] = None, cluster_marker: int = 0) -> None:
        if isinstance(padding_type, PaddingType):
            padding_type = padding_type.value
        self.padding_type_name: str = padding_type or PaddingType.UNPACKED.value
        self.lsb: bool = lsb
        self.alignment: Optional[int] = alignment
        self.cluster_marker: int = cluster_marker

        if self.padding_type_name not in _padding_type_names:
            raise ValueError(f'{self.padding_type_name} is not a valid padding type name.')

    @property
    def padding_type(self) -> PaddingType:
        return _padding_type_names[self.padding_type_name]

    def __eq__(self, other):
        return (isinstance(other, Packing) and self.padding_type == other.padding_type and
                self.lsb == other.lsb and self.alignment == other.alignment and
                self.cluster_marker == other.cluster_marker)

    def designation(self) -> str:
        lsb_string = '' if self.lsb else 'msb'
        padding_string: str = f'{self.padding_type_name}{lsb_string}'
        cluster_string = '' if not self.cluster_marker else f'c{self.cluster_marker}'

        alignment_string = ''
        if self.alignment:
            if self.padding_type is PaddingType.PACKED:
                padding_string += str(self.alignment)
            else:
                alignment_string = f'a{self.alignment}'

        return join_designators(join_designators(cluster_string, padding_string), alignment_string)

    @staticmethod
    def from_designation(full_designation: str) -> List[Tuple['Packing', str]]:

        possible_clusters: List[Tuple[int, str]] = [(0, full_designation)]
        if full_designation.startswith('c'):
            len_cluster_string: int = 0
            for len_cluster_string, c in full_designation[1:]:
                if not c.isdigit():
                    break

            if len_cluster_string and full_designation[1] != '0':
                possible_clusters.append((int(full_designation[1:len_cluster_string + 1]),
                                          full_designation[1 + len_cluster_string:]))

        possible_packing_types: List[Tuple[Packing, str]] = []

        for cluster, cluster_substring in possible_clusters:
            if cluster_substring.startswith('_'):
                cluster_substring = cluster_substring[1:]

            possible_paddings: List[Tuple[PaddingType, str]] = []
            for padding_type_name, padding_type in _padding_type_names.items():
                if cluster_substring[:len(padding_type_name)] == padding_type_name:
                    possible_paddings.append((padding_type, cluster_substring[len(padding_type_name):]))

            for padding, padding_substring in possible_paddings:
                possible_lsb: List[Tuple[bool, str]] = [(True, padding_substring)]
                if padding_substring.startswith('msb'):
                    possible_lsb.append((False, padding_substring[len('msb'):]))

                for lsb, lsb_substring in possible_lsb:
                    possible_alignments: List[Tuple[Optional[int], str]] = [(None, lsb_substring)]
                    alignment_possible: bool = False
                    if padding != PaddingType.PACKED:
                        if lsb_substring.startswith('a'):
                            lsb_substring = lsb_substring[1:]
                            alignment_possible = True
                    else:
                        alignment_possible = True

                    if alignment_possible:
                        len_alignment_int: int = 0
                        # @dev: We need to be able to iterate "past" the length of the string.
                        for len_alignment_int, c in enumerate(lsb_substring + 'X'):
                            if not c.isdigit():
                                break

                        if len_alignment_int:
                            possible_alignments.append((int(lsb_substring[:len_alignment_int]),
                                                        lsb_substring[len_alignment_int:]))

                    for alignment, alignment_substring in possible_alignments:
                        possible_packing_types.append((Packing(padding, lsb, alignment, cluster),
                                                       alignment_substring))

        return possible_packing_types
