from typing import Optional
from dataclasses import dataclass


@dataclass
class SensorModel:
    """Contains all information relevant to generic sensors."""


@dataclass(frozen=True)
class SensorModelDescriptor:
    camera_vendor: Optional[str]
    camera_model: Optional[str] = None
    camera_id: Optional[str] = None

    def __post_init__(self):
        if self.camera_id is not None:
            if self.camera_model is None or self.camera_vendor is None:
                raise ValueError('You need to specify vendor and model if you specify the '
                                 'camera identifier.')

        if self.camera_model is not None:
            if self.camera_vendor is None:
                raise ValueError('You need to specify vendor if you specify the camera identifier.')

    @property
    def description_level(self):
        if self.camera_id is not None:
            return 3

        if self.camera_model is not None:
            return 2

        if self.camera_vendor is not None:
            return 1

        return 0

    def score(self, camera_id: Optional[str], camera_model: Optional[str], camera_vendor: Optional[str]) -> int:
        description_level = self.description_level
        generality_score: int = -description_level

        if self.camera_id is not None and self.camera_id == camera_id:
            return 300 + generality_score

        if (self.camera_model is not None and camera_model is not None and
                camera_model.lower().startswith(self.camera_model.lower())):
            return 200 + generality_score

        if (self.camera_vendor is not None and camera_vendor is not None and
                self.camera_vendor.lower() == camera_vendor.lower()):
            return 100 + generality_score

        return generality_score

    def __str__(self) -> str:
        ident = '' if self.camera_id is None else f'{self.camera_id} '
        model = '' if self.camera_model is None else f'model "{self.camera_model}" '
        vendor = '' if self.camera_vendor is None else f'from vendor "{self.camera_vendor}"'
        descriptor = f'{ident}{model}{vendor}'
        return f'Camera Descriptor {descriptor}' if descriptor else 'Default Camera Descriptor'
