import copy
import time
import numpy as np

from abc import abstractmethod
from typing import TypeVar, Optional, Union, Dict, Tuple
from logging import getLogger

from dataspree.inference.core.item import Item
from dataspree.inference.sensor.data.camera import (
    Camera, CameraData, AcquisitionMode, TriggerType, TriggerMode, SoftwareTrigger, HardwareTrigger, TriggerActivation,
    TriggerAcquisitionMode, ContinuousAcquisitionMode, ManualGain, Roi, AutoGain, Brightness, TimedExposure,
    TriggeredExposure, AutoExposure, WhiteBalance, ManualWhiteBalance, AutoWhiteBalance, CheckerPatternRemoval,
    TonalRange, Binning, ContrastAdjustments, ColorAdjustment, TemperatureState, Gamma, NotAvailable)
from dataspree.inference.sensor.data.exceptions import (
    NoFreeHardwareTrigger, NoFreeSoftwareTrigger, InconsistentConfiguration, SensorException, SensorFeatureNotAvailable,
    FeatureNotImplemented, SensorFeatureCurrentlyNotAvailable, DriverState)
from dataspree.inference.sensor.data.pixel_format import PixelFormat, GenicamPixelFormat
from dataspree.inference.sensor.data.sensor import NotSupported
from dataspree.inference.sensor.driver.sensor_driver import SensorDriver

T = TypeVar('T', bound=Camera)

logger = getLogger(__name__)


# TODO: Think about error handling.

# Nomenclature:
# - Retrieve values that can be changed: fetch*
# - Update the value in the sensor configuration: refresh*
# - apply new values: set*
# - retrieve temporary values: retrieve*


class CameraDriver(SensorDriver[T]):
    def __init__(self):
        SensorDriver.__init__(self)

    def retrieve_effective_acquisition_rate(self, sensor: T) -> Union[float, NotAvailable]:
        """Return a camera's acquisition rate.

        Arguments:
            sensor (T): The sensor from which we gather the acquisition rate.

        Returns:
            Optional[float]: the camera's acquisition rate.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

        """

    def refresh_effective_acquisition_rate(self, sensor: T) -> None:
        """Fetch and update the effective acquisition rate.

        Arguments:
            sensor (T): The sensor from which we gather and update the frame rate.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

        See:
            self.fetch_effective_linerate
            self.fetch_effective_framerate

        """
        sensor = self._sensor(sensor)
        sensor.resulting_acquisition_rate = self.retrieve_effective_acquisition_rate(sensor)

    def refresh_attributes(self, sensor: T) -> None:
        """Refresh all available attributes.

        Arguments:
            sensor (T): the sensor for which the attributes should be refreshed

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

        """
        sensor = self._sensor(sensor)
        SensorDriver.refresh_attributes(self, sensor)

        self.refresh_acquisition_mode(sensor)
        self.refresh_operation_mode(sensor)
        self.refresh_pixel_format(sensor)
        self.refresh_current_camera_configuration_name(sensor)
        self.refresh_startup_camera_configuration_name(sensor)

        self.refresh_effective_acquisition_rate(sensor)

        self.refresh_crop(sensor)
        self.refresh_binning(sensor)

        self.refresh_brightness(sensor)
        self.refresh_white_balance(sensor)
        self.refresh_contrast_adjustments(sensor)
        self.refresh_color_adjustment(sensor)

        self.refresh_gamma(sensor)
        self.refresh_blooming_reduction(sensor)
        self.refresh_checker_pattern_removal(sensor)
        self.refresh_tonal_range_configuration(sensor)

        self.refresh_rois(sensor)

        self.refresh_camera_model(sensor)

    def set_attributes(self, sensor: T) -> None:
        """Apply all attributes that were provided.

        Fails when the first exception is thrown.

        Arguments:
            sensor (T): a sensor configuration that should be applied, that has all the fields.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound:
                If the sensor is currently not found.
            SensorStateException:
                If the sensor's state is not STARTED or CONNECTED.
            SensorStateInconsistent:
                If sensor connection objects are missing even though the state indicates they should be present.
            FeatureNotImplemented:
                If the provided `acquisition_mode` is not implemented in the `_apply_acquisition_mode` method.
        """
        old_configuration = self.sensors[sensor.identifier]

        if not old_configuration.acquisition_mode.entails(sensor.acquisition_mode):
            self.set_acquisition_mode(sensor, sensor.acquisition_mode)

        try:
            self.set_operation_mode(sensor, sensor.operation_mode)
        except (SensorException, FeatureNotImplemented, DriverState) as e:
            if not old_configuration.operation_mode.entails(sensor.operation_mode):
                raise e

        try:
            self.set_pixel_format(sensor, sensor.pixel_format)
        except (SensorException, FeatureNotImplemented, DriverState) as e:
            if not old_configuration.pixel_format.entails(sensor.pixel_format):
                raise e

        try:
            self.set_current_camera_configuration_name(sensor, sensor.current_camera_configuration)
        except (SensorException, FeatureNotImplemented, DriverState) as e:
            if not old_configuration.current_camera_configuration.entails(sensor.current_camera_configuration):
                raise e

        try:
            self.set_startup_camera_configuration_name(sensor, sensor.startup_camera_configuration)

        except (SensorException, FeatureNotImplemented, DriverState) as e:
            if not old_configuration.startup_camera_configuration.entails(sensor.startup_camera_configuration):
                raise e

        try:
            self.set_crop(sensor, sensor.crop)
        except (SensorException, FeatureNotImplemented, DriverState) as e:
            if not old_configuration.crop.entails(sensor.crop):
                raise e

        try:
            self.set_binning(sensor, sensor.binning)
        except (SensorException, FeatureNotImplemented, DriverState) as e:
            if not old_configuration.binning.entails(sensor.binning):
                raise e

        try:
            self.set_brightness(sensor, sensor.brightness)
        except (SensorException, FeatureNotImplemented, DriverState) as e:
            if not old_configuration.brightness.entails(sensor.brightness):
                raise e
        try:
            self.set_white_balance(sensor, sensor.white_balance)
        except (SensorException, FeatureNotImplemented, DriverState) as e:
            if not old_configuration.white_balance.entails(sensor.white_balance):
                raise e
        try:
            self.set_contrast_adjustments(sensor, sensor.contrast_adjustments)
        except (SensorException, FeatureNotImplemented, DriverState) as e:
            if not old_configuration.contrast_adjustments.entails(sensor.contrast_adjustments):
                raise e
        try:
            self.set_color_adjustment(sensor, sensor.color_adjustments)
        except (SensorException, FeatureNotImplemented, DriverState) as e:
            if not old_configuration.color_adjustments.entails(sensor.color_adjustments):
                raise e

        try:
            self.set_gamma(sensor, sensor.gamma)
        except (SensorException, FeatureNotImplemented, DriverState) as e:
            if not old_configuration.gamma.entails(sensor.gamma):
                raise e
        try:
            self.set_blooming_reduction(sensor, sensor.blooming_reduction)
        except (SensorException, FeatureNotImplemented, DriverState) as e:
            if not old_configuration.blooming_reduction.entails(sensor.blooming_reduction):
                raise e
        try:
            self.set_checker_pattern_removal(sensor, sensor.pattern_removal)
        except (SensorException, FeatureNotImplemented, DriverState) as e:
            if not old_configuration.pattern_removal.entails(sensor.pattern_removal):
                raise e
        try:
            self.set_tonal_range_configuration(sensor, sensor.tonal_range)
        except (SensorException, FeatureNotImplemented, DriverState) as e:
            if not old_configuration.tonal_range.entails(sensor.tonal_range):
                raise e

        try:
            self.set_rois(sensor, sensor.rois)
        except (SensorException, FeatureNotImplemented, DriverState) as e:
            if not old_configuration.rois.entails(sensor.rois):
                raise e

    def refresh_camera_model(self, sensor: T) -> None:
        """Refresh the camera model.

        Arguments:
            sensor (T): the sensor for which the attributes should be refreshed

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """

    #
    # Acquisition Mode
    #

    @abstractmethod
    def fetch_acquisition_mode(self, sensor: T) -> Union[AcquisitionMode, NotAvailable]:
        """Retrieve the currently active acquisition mode from the specified sensor.

        Args:
            sensor (T): The sensor object from which to retrieve the acquisition mode.

        Returns:
            AcquisitionMode: The current acquisition mode of the sensor.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound: If the specified sensor is not found.
            SensorStateException: If the sensor's state is neither `CONNECTED` nor `STARTED`.
            SensorStateInconsistent: If the sensor connection objects are missing despite the state indicating their
                presence.
        """

    def refresh_acquisition_mode(self, sensor: T) -> None:
        """Fetch and update the currently active acquisition mode from the specified sensor.

        Args:
            sensor (T): The sensor object from which to retrieve the acquisition mode.

        Returns:
            AcquisitionMode: The current acquisition mode of the sensor.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound: If the specified sensor is not found.
            SensorStateException: If the sensor's state is neither `CONNECTED` nor `STARTED`.
            SensorStateInconsistent: If the sensor connection objects are missing despite the state indicating their
                presence.
        """
        sensor = self._sensor(sensor)
        sensor.acquisition_mode = self.fetch_acquisition_mode(sensor)

    @abstractmethod
    def set_acquisition_mode(self, sensor: T, acquisition_mode: Union[AcquisitionMode, NotAvailable],
                             bounds_check: bool = False) -> None:
        """Set the acquisition mode for the specified sensor.

        Args:
            sensor (T): The sensor object for which the acquisition mode is to be set.
            acquisition_mode (AcquisitionMode): The acquisition mode to be applied to the sensor.
            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound:
                If the sensor is currently not found.
            SensorStateException:
                If the sensor's state is not STARTED or CONNECTED.
            SensorStateInconsistent:
                If sensor connection objects are missing even though the state indicates they should be present.
            FeatureNotImplemented:
                If the provided `acquisition_mode` is not implemented in the `_apply_acquisition_mode` method.

        Notes:
            - The method first tries to configure the frame rate settings if `max_rate` is specified.
            - If frame rate settings are not supported, it attempts to configure line rate settings.
            - The `_apply_acquisition_mode` method is used to apply additional acquisition settings.
            - Ensure that the `sensor` and `acquisition_mode` are correctly initialized and compatible with the method.
        """

    def set_acquisition_max_rate(self, sensor: T, max_acquisition_rate: Optional[Union[int, NotAvailable]]) -> None:
        """Set the maximum acquisition rate for the specified sensor.

        Args:
            sensor (T): The camera object for which the maximum acquisition rate is to be set.
            max_acquisition_rate (Optional[int]): The maximum acquisition rate to be applied. If `None`,
                                                  rate limitation is disabled.
        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound:
                If the sensor is currently not found.
            SensorStateException:
                If the sensor's state is not STARTED or CONNECTED.
            SensorStateInconsistent:
                If sensor connection objects are missing even though the state indicates they should be present.
            FeatureUnavailable:
                If the current  `acquisition_mode` is not implemented in the `_apply_acquisition_mode` method.
            AttributeError:
                If a required attribute cannot be set on the `image_acquirer`'s node map.
            AccessException:
                If there is an issue accessing or modifying specific attributes of the `image_acquirer`.
        """
        sensor = self._sensor(sensor)
        acquisition_mode = copy.deepcopy(sensor.acquisition_mode)
        acquisition_mode.max_rate = max_acquisition_rate
        self.set_acquisition_mode(sensor, acquisition_mode)

    def set_acquisition_continuous(self, sensor: T,
                                   max_acquisition_rate: Optional[Union[int, NotAvailable]] = None,
                                   bounds_check: bool = False) -> None:
        """Configure the specified sensor for continuous acquisition mode.

        Sets the acquisition mode of the given sensor to continuous acquisition. This mode captures frames continuously
        without acquisition triggers. Exposure triggers are not affected by this, and might still be required.

        Args:
            sensor (T): The camera object to be configured for continuous acquisition.

            max_acquisition_rate (Optional[int]): The maximum acquisition rate to be applied. If `None`,
                                                  rate limitation is disabled.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound:
                If the sensor is currently not found.
            SensorStateException:
                If the sensor's state is not STARTED or CONNECTED.
            SensorStateInconsistent:
                If sensor connection objects are missing even though the state indicates they should be present.
            AttributeError:
                If a required attribute cannot be set on the `image_acquirer`'s node map.
            AccessException:
                If there is an issue accessing or modifying specific attributes of the `image_acquirer`.
        """
        sensor = self._sensor(sensor)
        self.set_acquisition_mode(sensor, ContinuousAcquisitionMode(max_acquisition_rate),
                                  bounds_check=bounds_check)

    def set_acquisition_trigger_based(self, sensor: T, frame_start: TriggerType = TriggerType.SOFTWARE,
                                      frame_stop: Optional[Union[TriggerType, NotAvailable]] = None,
                                      max_acquisition_rate: Optional[Union[int, NotAvailable]] = None) -> None:
        """Configure the specified sensor for trigger-based acquisition mode.

        Sets the acquisition mode of the given sensor to trigger-based acquisition. Configures the triggers for starting
        and stopping frame capture based on the provided trigger types.

        Args:
            sensor (T): The camera object to be configured for trigger-based acquisition.
            frame_start (TriggerType, default=TriggerType.SOFTWARE): The trigger type for starting frame capture.
            frame_stop (Optional[TriggerType], default=None): The trigger type for stopping frame capture.
            max_acquisition_rate (Optional[int]): The maximum acquisition rate to be applied. If `None`,
                                                  rate limitation is disabled.

        Raises:
            DriverStateException: if the driver is in an invalid state

            NoFreeSoftwareTrigger: If there are no available software triggers to assign.
            NoFreeHardwareTrigger: If there are no available hardware triggers to assign.
            SensorNotFound:
                If the sensor is currently not found.
            SensorStateException:
                If the sensor's state is not STARTED or CONNECTED.
            SensorStateInconsistent:
                If sensor connection objects are missing even though the state indicates they should be present.
            AttributeError:
                If a required attribute cannot be set on the `image_acquirer`'s node map.
            AccessException:
                If there is an issue accessing or modifying specific attributes of the `image_acquirer`.
        """
        sensor = self._sensor(sensor)
        # Build Acquisition Mode:
        # 1. Get free triggers
        free_software_triggers = sensor.get_free_software_triggers()
        free_hardware_triggers = sensor.get_free_hardware_triggers()
        if isinstance(sensor.acquisition_mode, TriggerAcquisitionMode):
            for trigger in (sensor.acquisition_mode.trigger_frame_end, sensor.acquisition_mode.trigger_start):
                if sensor.is_software_trigger(trigger):
                    free_software_triggers.insert(0, trigger.trigger_source)
                elif sensor.is_hardware_trigger(sensor.acquisition_mode.trigger_frame_end):
                    free_hardware_triggers.insert(0, trigger.trigger_source)

        # 2. Set frame start trigger
        trigger_frame_start: TriggerMode
        if frame_start is TriggerType.SOFTWARE:
            try:
                trigger_name = sensor.camera_model.software_trigger_frame_start or free_software_triggers.pop()
                trigger_frame_start = SoftwareTrigger(trigger_name, 0)
            except IndexError:
                raise NoFreeSoftwareTrigger.create()
        else:
            try:
                trigger_frame_start = HardwareTrigger(free_hardware_triggers.pop(), 0, TriggerActivation.RISING_EDGE)
            except IndexError:
                raise NoFreeHardwareTrigger.create()

        # 3. Set frame stop trigger
        trigger_frame_stop: Optional[TriggerMode] = None
        if frame_stop is TriggerType.SOFTWARE:
            try:
                trigger_frame_stop = SoftwareTrigger(free_software_triggers.pop(), 0)
            except IndexError:
                raise NoFreeSoftwareTrigger.create()

        elif frame_stop is TriggerType.HARDWARE:
            trigger_frame_stop = HardwareTrigger(trigger_frame_start.trigger_source, 0, TriggerActivation.FALLING_EDGE)

        # 4. Either patch the trigger acquisition mode or initialize it, depending on the currently active mode.
        acquisition_mode: AcquisitionMode
        if not isinstance(sensor.acquisition_mode, TriggerAcquisitionMode):
            acquisition_mode = TriggerAcquisitionMode(max_acquisition_rate, trigger_start=trigger_frame_start,
                                                      trigger_frame_end=trigger_frame_stop)
        else:
            acquisition_mode = copy.copy(sensor.acquisition_mode)
            acquisition_mode.max_rate = max_acquisition_rate
            acquisition_mode.trigger_frame_start = trigger_frame_start
            acquisition_mode.trigger_frame_end = trigger_frame_stop

        # Attempt to apply the acquisition mode.
        if not sensor.acquisition_mode.entails(acquisition_mode):
            self.set_acquisition_mode(sensor, acquisition_mode)

    def set_acquisition_frame_burst(self, sensor: T, start: Optional[TriggerType] = TriggerType.SOFTWARE,
                                    stop: Optional[Union[TriggerType, NotAvailable]] = None,
                                    burst_frame_count: Optional[Union[int, NotAvailable]] = None,
                                    burst_mode: Optional[Union[str, NotAvailable]] = None) -> None:
        """Configure the sensor for frame burst acquisition mode.

        This method sets up the sensor to capture a series of frames in a burst mode, where each burst is triggered
        by specified start and stop triggers. It also allows for specifying the number of frames in the burst and the
        mode of burst acquisition.
        This method can only be used if the sensor is configured to run in trigger-based acquisition mode.

        Args:
            sensor (T): The camera object to be configured for frame burst acquisition.
            start (Optional[TriggerType], default=TriggerType.SOFTWARE): The trigger type to start the frame burst.
                Can be either `TriggerType.SOFTWARE` or `TriggerType.HARDWARE`.
            stop (Optional[TriggerType], default=None): The trigger type to stop the frame burst. Can be either
                `TriggerType.SOFTWARE` or `TriggerType.HARDWARE`. If `None`, the burst will not have a defined stop
                trigger.
            burst_frame_count (Optional[int], default=None): The number of frames to capture in each burst. If `None`,
                the frame count is not modified.
            burst_mode (Optional[str], default=None): The mode of burst acquisition, which may be specific to the camera
                model. If `None`, the burst mode is not modified.

        Raises:
            DriverStateException: if the driver is in an invalid state

            InconsistentConfiguration: If the sensor is not currently in trigger acquisition mode, which is required
                for frame burst acquisition.
            NoFreeSoftwareTrigger: If no available software triggers are found to assign.
            NoFreeHardwareTrigger: If no available hardware triggers are found to assign.
            SensorNotFound:
                If the sensor is currently not found.
            SensorStateException:
                If the sensor's state is not STARTED or CONNECTED.
            SensorStateInconsistent:
                If sensor connection objects are missing even though the state indicates they should be present.
            AttributeError:
                If a required attribute cannot be set on the `image_acquirer`'s node map.
            AccessException:
                If there is an issue accessing or modifying specific attributes of the `image_acquirer`.
        """
        sensor = self._sensor(sensor)

        if not isinstance(sensor.acquisition_mode, TriggerAcquisitionMode):
            raise InconsistentConfiguration('Frame burst acquisition can only be '
                                            'activated after activating trigger acquisition mode!')

        # Build Acquisition Mode:
        # 1. Get free triggers
        free_software_triggers = sensor.get_free_software_triggers()
        free_hardware_triggers = sensor.get_free_hardware_triggers()
        for trigger in (sensor.acquisition_mode.trigger_frame_burst_end,
                        sensor.acquisition_mode.trigger_frame_burst_start):
            if sensor.is_software_trigger(trigger):
                free_software_triggers.insert(0, trigger.trigger_source)
            elif sensor.is_hardware_trigger(sensor.acquisition_mode.trigger_frame_end):
                free_hardware_triggers.insert(0, trigger.trigger_source)

        # 2. Set start trigger
        trigger_frame_burst_start: Optional[TriggerMode] = None
        if start is TriggerType.SOFTWARE:
            try:
                trigger_frame_burst_start = SoftwareTrigger(free_software_triggers.pop(), 0)
            except IndexError:
                raise NoFreeSoftwareTrigger.create()
        elif start is TriggerType.HARDWARE:
            try:
                trigger_frame_burst_start = HardwareTrigger(free_hardware_triggers.pop(), 0,
                                                            TriggerActivation.RISING_EDGE)
            except IndexError:
                raise NoFreeHardwareTrigger.create()

        # 3. Set stop trigger
        trigger_frame_burst_stop: Optional[TriggerMode] = None
        if stop is TriggerType.SOFTWARE:
            try:
                trigger_frame_burst_stop = SoftwareTrigger(free_software_triggers.pop(), 0)
            except IndexError:
                raise NoFreeSoftwareTrigger.create()
        elif stop is TriggerType.HARDWARE:
            trigger_frame_burst_stop = HardwareTrigger(trigger_frame_burst_start.trigger_source, 0,
                                                       TriggerActivation.FALLING_EDGE)

        acquisition_mode = copy.copy(sensor.acquisition_mode)
        acquisition_mode.trigger_frame_burst_start = trigger_frame_burst_start
        acquisition_mode.trigger_frame_end = trigger_frame_burst_stop
        if burst_frame_count is not None:
            acquisition_mode.burst_frame_count = burst_frame_count
        if burst_mode is not None:
            acquisition_mode.burst_mode = burst_mode

        # Attempt to apply the acquisition mode.
        self.set_acquisition_mode(sensor, acquisition_mode)

    #
    # Acquisition status
    #

    def retrieve_acquisition_status_acquisition_active(self, sensor: T) -> bool:
        """Determines whether acquisition has started, i.e., whether the camera has received an
        Acquisition Start command.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Returns:
            bool: if the acquisition is currently active on the camera.

        Raises
            DriverStateException: if the driver is in an invalid state

            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If the status selector is not supported by the camera

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.
        """
        raise FeatureNotImplemented

    def retrieve_acquisition_status_acquisition_idle(self, sensor: T) -> bool:
        """Determine if the camera is currently idle.

        Arguments:
            sensor (T): The camera from which we gather the information.

        It is considered idle, if it is not currently exposing or reading out image data, and is therefore ready
        to receive an Acquisition Start command.

        Returns:
            bool: if the camera is currently idle.

        Raises
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If the status selector is not supported by the camera

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.
        """
        raise FeatureNotImplemented

    def retrieve_acquisition_status_exposure_active(self, sensor: T) -> bool:
        """Determines whether the camera is currently exposing.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Returns:
            bool: if the camera is currently exposing.

        Raises
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If the status selector is not supported by the camera

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.
        """
        raise FeatureNotImplemented

    def retrieve_acquisition_status_exposure_trigger_wait(self, sensor: T) -> bool:
        """Determines whether the camera is waiting for an Exposure Start trigger signal.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Returns:
            bool: if the camera is waiting for an Exposure Start trigger signal.

        Raises
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If the status selector is not supported by the camera

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.
        """
        raise FeatureNotImplemented

    def retrieve_acquisition_status_frame_burst_active(self, sensor: T) -> bool:
        """Determines whether the camera is currently acquiring a series of images (a "burst" of images).

        Arguments:
            sensor (T): The camera from which we gather the information.

        Returns:
            bool: if the camera is currently acquiring a series of images (a "burst" of images).

        Raises
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If the status selector is not supported by the camera

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.
        """
        raise FeatureNotImplemented

    def retrieve_acquisition_status_frame_burst_trigger_wait(self, sensor: T) -> bool:
        """Determines whether the camera is waiting for a Frame Burst Start trigger signal.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Returns:
            bool: if the camera is waiting for a Frame Burst Start trigger signal.

        Raises
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If the status selector is not supported by the camera

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.
        """
        raise FeatureNotImplemented

    def retrieve_acquisition_status_frame_trigger_wait(self, sensor: T) -> bool:
        """Determines whether the camera is waiting for a Frame Start trigger signal.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Returns:
            bool: if the camera is waiting for a Frame Start trigger signal.

        Raises
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If the status selector is not supported by the camera

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.
        """
        raise FeatureNotImplemented

    def retrieve_acquisition_status_line_trigger_wait(self, sensor: T) -> bool:
        """Determines whether the camera is waiting for a Line Start trigger signal.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Returns:
            bool: if the camera is waiting for a Line Start trigger signal.

        Raises:
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If the status selector is not supported by the camera

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.
        """
        raise FeatureNotImplemented

    #
    # Triggering
    #

    def execute_trigger(self, sensor: T, timestamp: Optional[int] = None) -> None:
        """Execute a trigger for acquiring data.

        This defaults to the frame start trigger.

        Arguments:
            sensor (T): The camera that is about to be triggered.

            timestamp (Optional[int]): Timestamp of the trigger

        Raises:
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If the status selector is not supported by the camera

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.

            InvalidArgumentException: Thrown if it is currently impossible to execute the request, for instance,
                                      because the camera is not in trigger acquisition mode.
        """
        sensor = self._sensor(sensor)
        self.execute_software_trigger_frame(sensor, True, timestamp=timestamp)

    @abstractmethod
    def execute_software_trigger_frame(self, sensor: T, start: bool, timestamp: Optional[int] = None) -> None:
        """Execute a software start or stop trigger for triggering a frame.

        Arguments:
            sensor (T): The camera that is about to be triggered.

            start (bool): frame start / frame stop.

            timestamp (Optional[int]): Timestamp of the trigger

        Raises:
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If the status selector is not supported by the camera

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.

            InvalidArgumentException: Thrown if it is currently impossible to execute the request, for instance,
                                      because the camera is not in trigger acquisition mode, or because the camera
                                      does not support / is not configured with a stop trigger.
        """

    def execute_software_trigger_frame_burst(self, sensor: T, start: bool) -> None:
        """Execute a software start or stop trigger for triggering a frame.

        Arguments:
            sensor (T): The camera that is about to be triggered.

            start (bool): frame start / frame stop.

        Raises:
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If the status selector is not supported by the camera

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.

            InvalidArgumentException: Thrown if it is currently impossible to execute the request, for instance,
                                      because the camera is not in trigger acquisition mode, or because the camera
                                      does not support / is not configured with a stop trigger.
        """
        raise FeatureNotImplemented

    def execute_software_trigger_exposure(self, sensor: T, start: bool) -> None:
        """Execute te software start or stop trigger that controls exposure.

        Arguments:
            sensor (T): The camera that is about to be triggered.

            start (bool): frame exposure / stop exposure.

        Raises:
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If the status selector is not supported by the camera

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.

            InvalidArgumentException: Thrown if it is currently impossible to execute the request, for instance,
                                      because the camera is not in trigger acquisition mode, or because the camera
                                      does not support / is not configured with a stop trigger.
        """
        raise FeatureNotImplemented

    def execute_software_trigger(self, sensor: T, trigger: Optional[Union[TriggerMode, str]],
                                 timestamp: Optional[int] = None) -> None:
        """Execute te software start or stop trigger that controls exposure.

        Arguments:
            sensor (T): The camera that is about to be triggered.

            trigger (Optional[Union[TriggerMode, str]]): Trigger that is supposed to be executed.

            timestamp (Optional[int]): Trigger that is supposed to be executed.

        Raises:
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If the status selector is not supported by the camera

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.

            InvalidArgumentException: Thrown if it is currently impossible to execute the request, for instance,
                                      because the camera is in an incompatible mode, or because the camera
                                      does not recognize / support the trigger that was provided.
        """
        raise FeatureNotImplemented

    def set_output_line_value(self, sensor: T, output_line_name: str, trigger_value: bool) -> None:
        """Set output line to high / low.

        Arguments:
            sensor (T): The camera that is about to be triggered.

            output_line_name (str): Name of the output line.

            trigger_value (bool): value the trigger should assume.

        Raises:
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If the status selector is not supported by the camera

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.

            InvalidArgumentException: Thrown if it is currently impossible to execute the request, for instance,
                                      because the camera is in an incompatible mode, or because the camera
                                      does not recognize / support the trigger that was provided.
        """
        raise FeatureNotImplemented

    #
    # Regions Of Interest
    #

    def highlight_roi(self, sensor: T, roi: str, highlight: bool) -> None:
        """Visually highlights a region of interest for configuration purpose.

        Arguments:
            sensor (T): The camera from which we gather the information.

            roi (str): the name of the region of interest

            highlight (bool) whether to highlight the ROI in the image.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is not available
        """
        raise FeatureNotImplemented

    def fetch_roi(self, sensor: T, roi: str) -> Union[Roi, NotAvailable]:
        """Retrieve roi from camera.

        Arguments:
            sensor (T): The camera from which we gather the information.

            roi (str): the name of the region of interest

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

        """
        return NotAvailable()

    def refresh_rois(self, sensor: T) -> None:
        """Fetch and update the camera's rois.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotImplemented if the feature is not available
        """
        sensor = self._sensor(sensor)
        sensor.rois = {roi: self.fetch_roi(sensor, roi) for roi in sensor.camera_model.rois}

    def set_roi(self, sensor: T, name: str, roi: Optional[Union[Roi, NotAvailable]]) -> None:
        """Configure an auto function ROI.

        The settings for the auto function rois are independent of the settings for the image roi.
        There should be no effect of the general ROI onto the configured auto rois.

        Attributes:
            sensor (T): The sensor, to which the roi settings are to be applied.

            roi (str): The name of the region of interest that is to be configured.

            roi: Optional[Roi]: The region of interest that is to be configured.
                                If this value is None, defaults to the entire image.

                                Attention: Don't use None if you configured a camera crop,
                                and want to use the same pixels as in the camera crop.
                                Use `sensor.crop` as argument instead!

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is not available

            InvalidArgumentException if the ranges are not valid.
        """
        raise FeatureNotImplemented

    def set_rois(self, sensor: T, rois: Dict[str, Union[Roi, Optional[NotAvailable]]]) -> None:
        """Configure auto function ROIs.

        Arguments:
            sensor (T): The camera from which we gather the information.

            rois (

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotImplemented if the feature is not available
        """
        sensor = self._sensor(sensor)
        for roi_name, roi_value in rois.items():
            self.set_roi(sensor, roi_name, roi_value)
        sensor.rois = rois

    #
    # Brightness
    #

    def fetch_brightness(self, sensor: T) -> Union[Brightness, NotAvailable]:
        """Return a camera's brightness settings.

        Returns:
            Optional[Brightness] the camera's brightness settings if they could be retrieved.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotImplemented if the feature is not available
        """
        return NotAvailable()

    def refresh_brightness(self, sensor: T) -> None:
        """Fetch and update the camera's brightness settings.

        Arguments:
            sensor (T): The camera from which we gather the information.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotImplemented if the feature is not available

        """
        sensor = self._sensor(sensor)
        sensor.brightness = self.fetch_brightness(sensor)
        if sensor.brightness.auto_function_enabled and sensor.brightness.auto_roi_configuration is None:
            logger.warning('Invalid brightness configuration: Auto functionality is enabled, but no region '
                           'of interest is provided.')

    def set_brightness(self, sensor: T, brightness: Optional[Union[Brightness, NotAvailable]],
                       bounds_check: bool = False):
        """Configure brightness.

        Arguments:
            sensor (T): The camera to which we apply the settings.

            brightness (Brightness) The brightness configuration.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If values are not applicable.

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.
        """
        raise FeatureNotImplemented

    #
    # Gain
    #

    def set_gain_manual(self, sensor: T, value: Union[float, int] = 0, digital: Optional[bool] = None,
                        bounds_check: bool = False) -> None:
        """Configure manual gain.

        Arguments:
            sensor (T): The camera to which we apply the gain settings.

            value: Union[float, int]: The configured gain. On some camera models, the gain must be entered as a "raw"
                                      value on an integer scale. The camera needs the raw value for its internal
                                      processing mechanism. The raw value, however, isn't the same as the actual gain
                                      value, which is expressed in decibels (dB).

            digital: Optional[bool]: Distinguishes between analog and digital gain.
                                     Analog gain is applied before the signal from the camera sensor is converted into
                                     digital values. Digital gain is applied after the conversion, i.e., it is basically
                                     a multiplication of the digitized values.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If values are not applicable.

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.
        """
        sensor = self._sensor(sensor)
        brightness = copy.deepcopy(sensor.brightness)
        brightness.gain = ManualGain(digital=digital, value=value)
        self.set_brightness(sensor, brightness, bounds_check=bounds_check)

    def set_gain_auto(self, sensor: T, upper_limit: Optional[Union[float, int]] = None,
                      lower_limit: Optional[Union[float, int]] = None,
                      auto_target_brightness: Optional[float] = None,
                      roi_configuration: Optional[Union[str, Roi]] = None,
                      digital: Optional[bool] = None, bounds_check: bool = False) -> None:
        """Configure automatic gain.

        Arguments:
            sensor (T): The camera to which we apply the gain settings.

            digital: Optional[bool]: Distinguishes between analog and digital gain.
                                     Analog gain is applied before the signal from the camera sensor is converted into
                                     digital values. Digital gain is applied after the conversion, i.e., it is basically
                                     a multiplication of the digitized values.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

            upper_limit: Union[float, int]: Upper limit for the gain. Raw integer value on a custom scale or
                                            actual gain value (float) expressed in decibels.
            lower_limit: Union[float, int]: Lower limit for the gain. Raw integer value on a custom scale or
                                            actual gain value (float) expressed in decibels.

            auto_target_brightness: Optional[float]: Brightness that is to be attained for automatic gain control.

            roi_configuration: Optional[Union[str, Roi]] Roi which is used to assess gain configuration.
                                                         If empty and the sensor supports roi configuration, use the
                                                         entire image as roi

        Raises:
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If values are not applicable.

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.
        """
        sensor = self._sensor(sensor)
        brightness = copy.deepcopy(sensor.brightness)
        brightness.gain = AutoGain(digital=digital, upper_limit=upper_limit, lower_limit=lower_limit, temporary=False)
        if auto_target_brightness:
            brightness.auto_target_brightness = auto_target_brightness
        if roi_configuration:
            brightness.auto_roi_configuration = roi_configuration
        self.set_brightness(sensor, brightness, bounds_check=bounds_check)

    def set_gain_manual_through_automatic_finder(
            self, sensor: T, upper_limit: Optional[Union[float, int]] = None,
            lower_limit: Optional[Union[float, int]] = None,
            auto_target_brightness: Optional[float] = None, roi_configuration: Optional[Union[str, Roi]] = None,
            digital: Optional[bool] = None, bounds_check: bool = False) -> None:
        """Run automatic gain once, and use the determined values for manual gain mode.

        Arguments:
            sensor (T): The camera to which we apply the gain settings.

            digital: Optional[bool]: Distinguishes between analog and digital gain.
                                     Analog gain is applied before the signal from the camera sensor is converted into
                                     digital values. Digital gain is applied after the conversion, i.e., it is basically
                                     a multiplication of the digitized values.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

            upper_limit: Union[float, int]: Upper limit for the gain. Raw integer value on a custom scale or
                                            actual gain value (float) expressed in decibels.
            lower_limit: Union[float, int]: Lower limit for the gain. Raw integer value on a custom scale or
                                            actual gain value (float) expressed in decibels.

            auto_target_brightness: Optional[float]: Brightness that is to be attained for automatic gain control.

            roi_configuration: Optional[Union[str, Roi]] Roi which is used to assess gain configuration.
                                                         If empty and the sensor supports roi configuration, use the
                                                         entire image as roi

        Raises:
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If values are not applicable.

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.
        """
        sensor = self._sensor(sensor)
        brightness = copy.deepcopy(sensor.brightness)
        brightness.gain = AutoGain(digital=digital, upper_limit=upper_limit, lower_limit=lower_limit, temporary=True)
        brightness.auto_target_brightness = auto_target_brightness
        brightness.auto_roi_configuration = roi_configuration
        self.set_brightness(sensor, brightness, bounds_check=bounds_check)

        time_start = time.time()
        brightness: Brightness
        while True:
            if (time.time() - time_start) >= 5.0:
                raise SensorFeatureCurrentlyNotAvailable('Timeout')
            brightness = self.fetch_brightness(sensor)
            if isinstance(brightness.gain, AutoGain):
                if not brightness.gain.temporary:
                    raise SensorException('Unexpected state retrieved.')
                time.sleep(0.5)
            elif brightness.gain is None:
                raise SensorFeatureNotAvailable
            else:
                break

        sensor.brightness = brightness

    #
    # Exposure
    #

    def set_exposure_timed(self, sensor: T, exposure_time: float, bounds_check: bool = False) -> None:
        """Configure timed exposure.

        Arguments:
            sensor (T): The camera to which we apply the settings.

            exposure_time (float) exposure time in microseconds.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If values are not applicable.

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.
        """
        sensor = self._sensor(sensor)
        brightness = copy.deepcopy(sensor.brightness)
        brightness.exposure = TimedExposure(exposure_time)
        self.set_brightness(sensor, brightness, bounds_check=bounds_check)

    def set_exposure_triggered(self, sensor: T, trigger_exposure_start: TriggerMode,
                               trigger_exposure_end: Optional[TriggerMode] = None,
                               start_delay: Optional[int] = None,
                               overlap_time_max: Optional[float] = None,
                               bounds_check: bool = False) -> None:
        """Configure triggered exposure.

        Arguments:
            sensor (T): The camera to which we apply the settings.

            trigger_exposure_start: Optional[TriggerMode] trigger that either indicates exposure start, or exposure
                                                          start and stop if the trigger has a
                                                          duration.

            trigger_exposure_end: Optional[TriggerMode] trigger that indicates stop of the exposure.

            start_delay: Optional[int]:  Determines the period of time between the detection of the trigger signal and
                                         the actual start of the exposure.

            overlap_time_max: Optional[float]: With overlapping image acquisition, exposure of the next image begins
                                               when the camera is reading out the sensor data of the previous image.
                                               In triggered exposure mode, the camera does not know how long the image
                                               is exposed. Overlap_time_max can be used to indicate the shortest
                                               exposure time that is intended to use be used in microseconds.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If values are not applicable.

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.
        """
        sensor = self._sensor(sensor)

        brightness = copy.deepcopy(sensor.brightness)
        brightness.exposure = TriggeredExposure(trigger_exposure_start=trigger_exposure_start,
                                                trigger_exposure_end=trigger_exposure_end,
                                                start_delay=start_delay, overlap_time_max=overlap_time_max)
        self.set_brightness(sensor, brightness, bounds_check=bounds_check)

    def set_exposure_auto(self, sensor: T, upper_limit: Optional[Union[float, int]] = None,
                          lower_limit: Optional[Union[float, int]] = None,
                          auto_target_brightness: Optional[float] = None,
                          roi_configuration: Optional[Union[str, Roi]] = None, bounds_check: bool = False) -> None:
        """Configure automatic gain.

        Arguments:
            sensor (T): The camera to which we apply the gain settings.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

            upper_limit: Union[float, int]: Upper limit for the gain. Raw integer value on a custom scale or
                                            actual gain value (float) expressed in decibels.
            lower_limit: Union[float, int]: Lower limit for the gain. Raw integer value on a custom scale or
                                            actual gain value (float) expressed in decibels.

            auto_target_brightness: Optional[float]: Brightness that is to be attained for automatic gain control.

            roi_configuration: Optional[Union[str, Roi]] Roi which is used to assess gain configuration.
                                                         If empty and the sensor supports roi configuration, use the
                                                         entire image as roi

        Raises:
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If values are not applicable.

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.
        """
        sensor = self._sensor(sensor)
        brightness = copy.deepcopy(sensor.brightness)
        brightness.exposure = AutoExposure(upper_limit=upper_limit, lower_limit=lower_limit, temporary=False)
        if auto_target_brightness:
            brightness.auto_target_brightness = auto_target_brightness
        if roi_configuration:
            brightness.auto_roi_configuration = roi_configuration
        self.set_brightness(sensor, brightness, bounds_check=bounds_check)

    def set_exposure_manual_through_automatic_finder(self, sensor: T, upper_limit: Optional[Union[float, int]] = None,
                                                     lower_limit: Optional[Union[float, int]] = None,
                                                     auto_target_brightness: Optional[float] = None,
                                                     roi_configuration: Optional[Union[str, Roi]] = None,
                                                     bounds_check: bool = False) -> None:
        """Run automatic exposure once, and use the determined values for timed exposure mode.

        Arguments:
            sensor (T): The camera to which we apply the gain settings.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

            upper_limit: Union[float, int]: Upper limit for the gain. Raw integer value on a custom scale or
                                            actual gain value (float) expressed in decibels.
            lower_limit: Union[float, int]: Lower limit for the gain. Raw integer value on a custom scale or
                                            actual gain value (float) expressed in decibels.

            auto_target_brightness: Optional[float]: Brightness that is to be attained for automatic gain control.

            roi_configuration: Optional[Union[str, Roi]] Roi which is used to assess gain configuration.
                                                         If empty and the sensor supports roi configuration, use the
                                                         entire image as roi

        Raises:
            DriverStateException: if the driver is in an invalid state

            InvalidArgumentException If values are not applicable.

            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.
        """
        sensor = self._sensor(sensor)
        brightness = copy.deepcopy(sensor.brightness)
        brightness.exposure = AutoExposure(upper_limit=upper_limit, lower_limit=lower_limit, temporary=True)
        if auto_target_brightness:
            brightness.auto_target_brightness = auto_target_brightness
        if roi_configuration:
            brightness.auto_roi_configuration = roi_configuration
        self.set_brightness(sensor, brightness, bounds_check=bounds_check)

        brightness: Brightness
        time_start = time.time()
        while True:
            if (time.time() - time_start) >= 5.0:
                raise SensorFeatureCurrentlyNotAvailable('Timeout')
            brightness = self.fetch_brightness(sensor)
            if isinstance(brightness.exposure, AutoExposure):
                if not brightness.exposure.temporary:
                    raise SensorException('Unexpected state retrieved.')
                time.sleep(0.5)
            elif isinstance(brightness.exposure, TimedExposure):
                break
            else:
                raise SensorFeatureNotAvailable

        sensor.brightness = brightness

        self.refresh_effective_acquisition_rate(sensor)

    #
    # White balance
    #
    def fetch_white_balance(self, sensor) -> Union[WhiteBalance, NotAvailable]:
        """Return the white balance settings

        Arguments:
            sensor (T): The sensor from which we gather the white balance.

        Returns:
            Optional[float]: the camera's white balance.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotImplemented if the feature is not implemented.

        """
        return NotAvailable()

    def refresh_white_balance(self, sensor: T) -> None:
        """Retrieve and apply the white balance settings

        Arguments:
            sensor (T): The sensor from which we gather the white balance.

        Returns:
            Optional[float]: the camera's white balance.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotImplemented if the feature is not implemented.

        """
        sensor = self._sensor(sensor)
        sensor.white_balance = self.fetch_white_balance(sensor)

    def set_white_balance(self, sensor: T, white_balance: Union[WhiteBalance, NotAvailable],
                          bounds_check: bool = False) -> None:
        """Apply the white balance settings.

        Arguments:
            sensor (T): The sensor to which apply the white balance settings

            white_balance (WhiteBalance): The white balance settings

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.

        """
        raise FeatureNotImplemented

    def set_white_balance_manual(self, sensor: T, balance_ratios: Optional[Dict[str, float]] = None,
                                 bounds_check: bool = False) -> None:
        """Apply manual white balance

        Arguments:
            sensor (T): The sensor to which apply the white balance settings

            balance_ratios (Dict[str, float]): Contains the balance ratio for each channel. Typically,
                the balance ratios are configured as follows:
                balance_ratio == 1: The intensity of the color remains unchanged.
                               > 1: The intensity of the color is increased relative to the other two colors.
                               < 1: The intensity of the color is decreased relative to the other two colors.
                However, some cameras don't allow balance ratios smaller than one, because a balance
                can be normalized to an equivalent balance by dividing through the sum of all
                balance ratios or multiplying each value with a specific factor.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.

        """
        # TODO: we could think about replacing 1.0 by the sum (?) of all values.
        #       This also needs to be tested with a RGB camera.
        sensor = self._sensor(sensor)
        full_ratios = {channel: 1.0 for channel in sensor.camera_model.white_balance_manual_channels}
        full_ratios.update(balance_ratios or dict())
        self.set_white_balance(sensor, ManualWhiteBalance(full_ratios), bounds_check=bounds_check)

    def set_white_balance_auto(self, sensor: T, roi_configuration: Optional[Union[str, Roi]] = None,
                               adjustment_damping: Optional[float] = None, bounds_check: bool = False) -> None:
        """Configure automatic white balance finder.

        Arguments:
            sensor (T): The sensor to which apply the white balance settings

            roi_configuration (Optional[Union[str, Roi]]): The roi on which the auto white balance is executed

            adjustment_damping (Optional[float]):  If specified, colors are not adjusted immediately, but after a
                certain delay. This can be useful to avoid the automatic control mechanism of the auto functions being
                disrupted by objects moving in and out of the camera's area of view.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                Otherwise, this function raises if we f.i. attempt to configure a gain value of 1000.0 even though the
                maximum allowed gain is 100.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.
        """
        sensor = self._sensor(sensor)
        self.set_white_balance(sensor, AutoWhiteBalance(roi_configuration, adjustment_damping, temporary=False),
                               bounds_check=bounds_check)

    def set_white_balance_manual_through_automatic_finder(self, sensor: T,
                                                          roi_configuration: Optional[Union[str, Roi]] = None,
                                                          adjustment_damping: Optional[float] = None,
                                                          bounds_check: bool = False) -> None:
        """Run automatic white balance finder once, and thus determine and use parameters for manual white balance mode.

        Arguments:
            sensor (T): The sensor to which apply the white balance settings

            roi_configuration (Optional[Union[str, Roi]]): The roi on which the auto white balance is executed

            adjustment_damping (Optional[float]):  If specified, colors are not adjusted immediately, but after a
                certain delay. This can be useful to avoid the automatic control mechanism of the auto functions being
                disrupted by objects moving in and out of the camera's area of view.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                Otherwise, this function raises if we f.i. attempt to configure a gain value of 1000.0 even though the
                maximum allowed gain is 100.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.
        """
        sensor = self._sensor(sensor)

        self.set_white_balance(sensor, AutoWhiteBalance(roi_configuration, adjustment_damping, temporary=True),
                               bounds_check=bounds_check)

        time_start = time.time()
        white_balance: WhiteBalance
        while True:
            if (time.time() - time_start) >= 5.0:
                raise SensorFeatureCurrentlyNotAvailable('Timeout')
            white_balance = self.fetch_white_balance(sensor)
            if isinstance(white_balance, AutoWhiteBalance):
                if not white_balance.temporary:
                    raise SensorException('Unexpected auto white balance state retrieved.')
                time.sleep(0.5)
            elif isinstance(white_balance, ManualWhiteBalance):
                break
            else:
                raise SensorFeatureCurrentlyNotAvailable

        sensor.white_balance = white_balance

    #
    # Checker pattern removal
    #

    def fetch_checker_pattern_removal(self, sensor) -> Union[CheckerPatternRemoval, NotAvailable]:
        """Return the checker pattern removal settings.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Optional[CheckerPatternRemoval]: the camera's checker pattern removal configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

        """
        return NotAvailable()

    def refresh_checker_pattern_removal(self, sensor: T) -> None:
        """Retrieve and apply the checker pattern removal settings.

        Arguments:
            sensor (T): The sensor from which we gather the checker pattern removal.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotImplemented if the feature is not implemented.
        """
        sensor = self._sensor(sensor)
        sensor.pattern_removal = self.fetch_checker_pattern_removal(sensor)

    def set_checker_pattern_removal(self, sensor: T, checker_pattern_removal: CheckerPatternRemoval,
                                    bounds_check: bool = False) -> None:
        """Apply the checker pattern removal settings.

        Arguments:
            sensor (T): The sensor to which apply the checker pattern removal settings.

            checker_pattern_removal (CheckerPatternRemoval): The checker pattern removal settings.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.

        """
        raise FeatureNotImplemented

    #
    # Tonal Range Configuration
    #

    def fetch_tonal_range_configuration(self, sensor) -> Union[TonalRange, NotAvailable]:
        """Return the tonal range configuration.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Optional[TonalRange]: the camera's tonal range configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """
        return NotAvailable()

    def refresh_tonal_range_configuration(self, sensor: T) -> None:
        """Retrieve and apply the tonal range configuration.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotImplemented if the feature is not implemented.
        """
        sensor = self._sensor(sensor)
        sensor.tonal_range = self.fetch_tonal_range_configuration(sensor)

    def set_tonal_range_configuration(self, sensor: T, checker_pattern_removal: Union[TonalRange, NotAvailable],
                                      bounds_check: bool = False) -> None:
        """Apply the checker pattern removal settings.

        Arguments:
            sensor (T): The sensor to which apply the checker pattern removal settings.

            checker_pattern_removal (CheckerPatternRemoval): The checker pattern removal settings.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.

        """
        raise FeatureNotImplemented

    #
    # Binning
    #

    def fetch_binning(self, sensor: T) -> Union[Binning, NotAvailable]:
        """Return the binning configuration.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Optional[Binning]: the camera's binning configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotImplemented if the feature is not implemented.
        """
        return NotAvailable()

    def refresh_binning(self, sensor: T) -> None:
        """Retrieve and apply the binning configuration.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotImplemented if the feature is not implemented.
        """
        sensor = self._sensor(sensor)
        sensor.binning = self.fetch_binning(sensor)

    def set_binning(self, sensor: T, binning: Optional[Binning] = None, bounds_check: bool = False) -> None:
        """Apply the binning configuration.

        Arguments:
            sensor (T): The sensor to which apply the binning configuration.

            binning (Optional[Binning]): The binning configuration or None to disable binning.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.

        """
        raise FeatureNotImplemented

    #
    # Blooming reduction
    #
    def fetch_blooming_reduction(self, sensor: T) -> Union[bool, NotAvailable]:
        """Return the blooming reduction configuration.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Optional[bool]: the T's blooming reduction configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotImplemented if the feature is not implemented.
        """
        return NotAvailable()

    def refresh_blooming_reduction(self, sensor: T) -> None:
        """Retrieve and apply the blooming reduction configuration.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotImplemented if the feature is not implemented.
        """
        sensor = self._sensor(sensor)
        sensor.blooming_reduction = self.fetch_blooming_reduction(sensor)

    def set_blooming_reduction(self, sensor: T, blooming_reduction: bool) -> None:
        """Apply the blooming reduction configuration.

        Arguments:
            sensor (T): The sensor to which apply the blooming reduction configuration.

            blooming_reduction (bool): The blooming reduction configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.
        """
        raise FeatureNotImplemented

    #
    # Contrast Adjustments
    #

    def fetch_contrast_adjustments(self, sensor: T) -> Union[ContrastAdjustments, NotAvailable]:
        """Return the contrast adjustment configuration.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Optional[bool]: the T's contrast adjustment configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotImplemented if the feature is not implemented.
        """
        return NotAvailable()

    def refresh_contrast_adjustments(self, sensor: T) -> None:
        """Retrieve and apply the contrast adjustment configuration.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotImplemented if the feature is not implemented.
        """
        sensor = self._sensor(sensor)
        sensor.contrast_adjustments = self.fetch_contrast_adjustments(sensor)

    def set_contrast_adjustments(self, sensor: T, contrast_adjustments: ContrastAdjustments,
                                 bounds_check: bool = False) -> None:
        """Apply the contrast adjustment configuration.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

            contrast_adjustments (bool): The contrast adjustment configuration.

            bounds_check (bool): If true, adjusts the number values to adhere to the bounds required from the sensor.
                                 Otherwise, this function raises if we f.i. attempt to configure a gain value of
                                 1000.0 even though the maximum allowed gain is 100.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.
        """
        raise FeatureNotImplemented

    #
    # Operation mode
    #

    def fetch_operation_mode(self, sensor: T) -> Union[str, NotAvailable]:
        """Return the operation mode.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Optional[bool]: The sensor's operation mode.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotImplemented if the feature is not implemented.
        """
        return NotAvailable()

    def refresh_operation_mode(self, sensor: T) -> None:
        """Retrieve and apply the operation mode.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotImplemented if the feature is not implemented.
        """
        sensor = self._sensor(sensor)
        sensor.operation_mode = self.fetch_operation_mode(sensor)

    def set_operation_mode(self, sensor: T, operation_mode: Union[str, NotAvailable]) -> None:
        """Apply the operation mode.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

            operation_mode (str): the new operation mode

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.
        """
        raise FeatureNotImplemented

    #
    # Color adjustment
    #

    def fetch_color_adjustment(self, sensor: T) -> Union[ColorAdjustment, NotAvailable]:
        """Return the color adjustment.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Optional[bool]: The sensor's color adjustment.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotImplemented if the feature is not implemented.
        """
        return NotAvailable()

    def refresh_color_adjustment(self, sensor: T) -> None:
        """Retrieve and apply the color adjustment.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotImplemented if the feature is not implemented.
        """
        sensor = self._sensor(sensor)
        sensor.color_adjustments = self.fetch_color_adjustment(sensor)

    def set_color_adjustment(self, sensor: T, color_adjustment: ColorAdjustment) -> None:
        """Apply the color adjustment.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

            color_adjustment (ColorAdjustment): the new color adjustment.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.
        """
        raise FeatureNotImplemented

    #
    # Gamma
    #

    def fetch_gamma(self, sensor: T) -> Union[Gamma, NotAvailable]:
        """Return the gamma settings.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Optional[bool]: The sensor's gamma configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """
        return NotAvailable()

    def refresh_gamma(self, sensor: T) -> None:
        """Retrieve and apply the gamma configuration.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotImplemented if the feature is not implemented.
        """
        sensor = self._sensor(sensor)
        sensor.gamma = self.fetch_gamma(sensor)

    def set_gamma(self, sensor: T, gamma: Gamma) -> None:
        """Apply the gamma configuration.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

            gamma (Gamma): the new gamma configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.
        """
        raise FeatureNotImplemented

    #
    # Device temperature
    #

    def retrieve_device_temperature(self, sensor: T) -> Union[float, NotAvailable]:
        """Return the current device temperature.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Optional[float]: The sensor's temperature.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotImplemented if the feature is not implemented.
        """
        raise FeatureNotImplemented

    def retrieve_device_temperature_state(self, sensor: T) -> Union[TemperatureState, NotAvailable]:
        """Return the current device temperature state.

        The Temperature State camera indicates whether the camera's internal temperature is normal or too high.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Optional[float]: The sensor's temperature state

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotImplemented if the feature is not implemented.
        """
        raise FeatureNotImplemented

    #
    # Pixel format
    #
    @abstractmethod
    def fetch_pixel_format(self, sensor: T) -> Union[PixelFormat, NotAvailable]:
        """Return the current device pixel format.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Optional[float]: The sensor's pixel format.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """

    def refresh_pixel_format(self, sensor: T) -> None:
        """Retrieve and apply the pixel format.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """
        sensor = self._sensor(sensor)
        sensor.pixel_format = self.fetch_pixel_format(sensor)

    @abstractmethod
    def set_pixel_format(self, sensor: T, pixel_format: PixelFormat) -> None:
        """Apply the pixel format.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

            pixel_format (PixelFormat): the new pixel format.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.
        """

    #
    # Crop
    #

    def fetch_crop(self, sensor: T) -> Union[Roi, NotAvailable]:
        """Return the current crop.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Optional[Roi]: The sensor's crop.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """
        return NotSupported()

    def refresh_crop(self, sensor: T) -> None:
        """Retrieve and apply the crop.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """
        sensor = self._sensor(sensor)
        sensor.crop = self.fetch_crop(sensor)

    def set_crop(self, sensor: T, crop: Optional[Roi] = None) -> None:
        """Apply the crop.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

            crop (Roi): the new crop.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.

            InvalidArgumentException if the roi cannot be applied, because the values are invalid.
        """
        raise FeatureNotImplemented

    #
    # Camera configurations
    #

    def fetch_current_camera_configuration_name(self, sensor: T) -> Union[str, NotAvailable]:
        """Return the current camera configuration.

        This does not convey any information on whether the current settings are actually loaded from the camera
        configuration.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Optional[str]: The sensor's camera configuration name.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """

    def refresh_current_camera_configuration_name(self, sensor: T) -> None:
        """Retrieve and apply the current camera configuration name.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """
        sensor = self._sensor(sensor)
        sensor.current_camera_configuration = self.fetch_current_camera_configuration_name(sensor)

    def set_current_camera_configuration_name(self, sensor: T, current_camera_configuration_name: str) -> None:
        """Sent the current camera configuration to the camera configuration that is identified
        by the provided name. This does not automatically load the configuration, as we might
        want to set the current camera configuration just to apply our current configuration to it.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

            current_camera_configuration_name (str): the camera configuration to apply / load.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.

            InvalidArgumentException if the roi cannot be applied, because the values are invalid.
        """
        raise FeatureNotImplemented

    def load_current_camera_configuration(self, sensor: T) -> None:
        """Load the current configuration.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.

            InvalidArgumentException if the roi cannot be applied, because the values are invalid.
        """
        raise FeatureNotImplemented

    def save_current_camera_configuration(self, sensor: T) -> None:
        """Save the current configuration.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.

            InvalidArgumentException if the roi cannot be applied, because the values are invalid.
        """
        raise FeatureNotImplemented

    def fetch_startup_camera_configuration_name(self, sensor: T) -> Union[str, NotAvailable]:
        """Return the startup camera configuration name.

        This does not convey any information on whether the current settings are actually loaded from the camera
        configuration.

        Arguments:
            sensor (T): The sensor from which we gather information.

        Returns:
            Optional[str]: The sensor's camera configuration name.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """
        return NotSupported()

    def refresh_startup_camera_configuration_name(self, sensor: T) -> None:
        """Retrieve and apply the camera startup configuration name.

        This does not convey any information on whether the current settings are actually loaded from the camera
        configuration.

        Arguments:
            sensor (T): The sensor to which we apply the configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """
        sensor = self._sensor(sensor)
        sensor.startup_camera_configuration = self.fetch_startup_camera_configuration_name(sensor)

    def set_startup_camera_configuration(self, sensor: T, startup_camera_configuration_name: str) -> None:
        """Sent the startup camera configuration to the camera configuration that is identified
        by the provided name. This does not automatically load the configuration, as we might
        want to set the current camera configuration just to apply our current configuration to it.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

            startup_camera_configuration_name (str): the camera configuration to apply / load.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            InvalidArgumentException if the roi cannot be applied, because the values are invalid.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.

        """
        raise FeatureNotImplemented

    def load_startup_camera_configuration(self, sensor: T) -> None:
        """Load the configured startup configuration.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.

            InvalidArgumentException if the roi cannot be applied, because the values are invalid.
        """
        sensor = self._sensor(sensor)
        self.set_current_camera_configuration_name(sensor, sensor.startup_camera_configuration)
        self.load_current_camera_configuration_name(sensor)

    def save_current_configuration_as_startup_configuration(
            self, sensor: T, startup_camera_configuration_name: Optional[str] = None) -> None:
        """Save the current configuration as startup configuration.

        Arguments:
            sensor (T): The sensor to which apply the configuration.

            startup_camera_configuration_name (Optional[str]): optionally provide a new startup configuration name.
                                                               If None, use the currently configured startup config.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.

            InvalidArgumentException if the roi cannot be applied, because the values are invalid.
        """
        sensor = self._sensor(sensor)
        if startup_camera_configuration_name is not None:
            self.set_startup_camera_configuration(sensor, startup_camera_configuration_name)

        self.set_current_camera_configuration_name(sensor, sensor.startup_camera_configuration)
        self.save_current_camera_configuration(sensor)

    #
    # Data acquisition
    #

    @abstractmethod
    def fetch_sensor_data(self, sensor: T, timeout: float = 0.0) -> Union[CameraData, NotAvailable]:
        ...

    def create_item(self, sensor: T, sensor_data: CameraData) -> Item:
        sensor = self._sensor(sensor)
        pixel_format = sensor.pixel_format

        try:
            image_structured = pixel_format.get_structured_array(sensor_data.content)
        except Exception as e:
            logger.warning(e)
            image_structured = None

        image: Optional[np.ndarray] = None
        if isinstance(pixel_format, GenicamPixelFormat) and len(sensor_data.content) > 1:
            channel_designation = pixel_format.component.channel_designation()

            # Specific presets.
            channel_designations: Dict[str, Tuple[int, ...]] = {
                'RGB': (0, 1, 2),
                'YRGB': (0, 1, 2, 3),
                'RGBa': (0, 1, 2, 3),
                'aRGB': (0, 1, 2, 3),
                'HSI': (0, 1, 2),
                'HSV': (0, 1, 2),
                'YCbCr': (0, 1, 2),
                'aBGR': (0, 1, 2, 3),
                'BGRa': (0, 1, 2, 3),
            }

            designation = channel_designations.get(channel_designation, tuple((range(len(pixel_format.component)))))

            fallback = not len(designation)
            for idx in designation:
                if 0 <= idx < len(sensor_data.content):
                    first = sensor_data.content[designation[0]]
                    current = sensor_data.content[idx]
                    fallback = not (current.shape == first.shape) or not (current.dtype == first.dtype)
                else:
                    fallback = True

                if fallback:
                    break

            if not fallback:

                first = sensor_data.content[designation[0]]

                image = np.empty((*first.shape, len(designation)), first.dtype)
                for i, idx in enumerate(designation):
                    image[..., i] = sensor_data.content[idx]

                if image.shape[-2] == 1:
                    image = image[..., 0, :]

        if image is None:
            image = sensor_data.content[0] if len(sensor_data.content) else None

        try:
            sensor_copy = copy.deepcopy(sensor)
        except Exception:
            sensor_copy = None

        trigger_timestamp = sensor.trigger_timestamp
        sensor.trigger_timestamp = None

        return Item.default(id=f'{sensor.identifier}_',
                            image=image,
                            image_structured=image_structured,
                            image_channels=sensor_data.content,
                            camera_identifier=sensor.identifier,
                            camera_model=sensor.camera_model,
                            camera_manufacturer=sensor.manufacturer,
                            camera_name=sensor.name,
                            sensor=sensor_copy,
                            gain=sensor_data.gain,
                            exposure_time=sensor_data.exposure_time,
                            sensor_data=sensor_data,
                            trigger_count=sensor.trigger_count,
                            timestamp_trigger=trigger_timestamp,
                            timestamp_acquisition=sensor.acquisition_timestamp)
