from abc import abstractmethod, ABCMeta
from threading import Condition
from typing import Generic, Optional, TypeVar, Dict, List, Union

from dataspree.inference.sensor.data import SensorState
from dataspree.inference.sensor.data.sensor import Sensor, SensorData, SensorCallback, NotAvailable

from dataspree.inference.sensor.data.exceptions import FeatureNotImplemented, DriverState, SensorNotFound

T = TypeVar('T', bound=Sensor)


class Sensors(Generic[T]):

    def __init__(self, cb: SensorCallback):
        self.__cb: SensorCallback = cb
        self.__values: Dict[str, Sensor] = {}

    def add_sensor(self, sensor: T):
        if sensor.identifier in self.__values:
            raise ValueError('Sensor already added.')

        if sensor.state != SensorState.DISCOVERED:
            raise ValueError

        sensor.register_callback(self.__cb)

        self.__values[sensor.identifier] = sensor
        self.__cb.on_sensor_tracked(sensor.identifier)
        self.__cb.on_sensor_found(sensor)

    def __iter__(self):
        return self.__values.__iter__()

    def __len__(self):
        return self.__values.__len__()

    def __getitem__(self, item):
        return self.__values.__getitem__(item)

    def get(self, value, default=None):
        return self.__values.get(value, default)

    def items(self):
        return self.__values.items()

    def values(self):
        return self.__values.values()

    def keys(self):
        return self.__values.keys()


class SensorDriver(Generic[T], SensorCallback[T], metaclass=ABCMeta):
    """

    TODO: add some general documentation

    TODO: Here, I describe the uniform properties on all the functions that are

    - fetch_*
    - refresh_*
    - set_*

    """

    def __init__(self) -> None:
        Generic.__init__(self)
        self.__sensors: Optional[Sensors[T]] = None

        # Can be notified by the sensor driver when data is available.
        # TODO: Document usage, check that this is actually used!
        self.condition_variable = Condition()
        self.__listeners: List[SensorCallback[T]] = []

    @property
    def sensors(self) -> Sensors[T]:
        sensors = self.__sensors
        if sensors is None:
            raise DriverState()
        return sensors

    def destroy_sensors(self):
        ex = None
        for sens in self.__sensors.values():
            try:
                sens.state = SensorState.UNAVAILABLE
            except Exception as e:
                ex = e

        self.__sensors = None
        if ex is not None:
            raise ex from None

    def create_sensors(self):
        self.__sensors = Sensors[T](self)

    @abstractmethod
    def setup(self) -> 'SensorDriver[T]':
        """Initialize the driver. This function must be called before any other functionality can be used.

        Initializes the list of sensors and driver internals.

        Raises:
            InvalidState if the driver appears to already be set up.
        """

    @abstractmethod
    def destroy(self) -> None:
        """Destroys the driver. After calling destroy, no other function can be called successfully aside from setup.

        Destroys the list of sensors and unregisters driver internals.

        Raises:
            InvalidState if the driver is not properly set up.
        """

    @abstractmethod
    def retrieve_sensors(self) -> None:
        """Retrieve all available sensors and update self.sensors.

        Does not refresh sensor attributes.

        Raises:
            InvalidState if the driver is not properly set up.
        """

    def reset_driver(self) -> None:
        """Reset the driver.

        Temporarily disconnects and stops all cameras. This function can be called at any point in time.

        1. Driver internals are reset
        2. The driver stops and disconnects all sensors.
        3. Call to `destroy`
        4. Call to `setup`
        5. Call to `retrieve_sensors`
        6. The driver reconnects to all previously connected sensors and starts all previously started sensors.

        Raises:
            SensorError if reconnecting to some devices failed.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.
        """
        raise FeatureNotImplemented

    def __enter__(self):
        """Initialize the driver. This function must be called before any other functionality can be used.

        Initializes the list of sensors and driver internals.

        Raises:
            InvalidState if the driver appears to already be set up.
        """
        return self.setup()

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Destroys the driver. After calling destroy, no other function can be called successfully aside from setup.

        Destroys the list of sensors and unregisters driver internals.

        Raises:
            InvalidState if the driver is not properly set up.
        """
        self.destroy()

    #
    # Sensor state
    #

    @abstractmethod
    def connect(self, sensor: T) -> None:
        """Connect to the provided sensor.

        This creates the necessary internal connection objects.
        After the connection is established, the sensor's attributes are fetched and stored in the sensor instance.

        Arguments:
            sensor (T) The sensor which is supposed to be connected.

        Raises:
            SensorNotFound if the sensor does not exist.

            SensorStateException if the sensor's state is not DISCOVERED

        See:
            `self.refresh_attributes`: Function used for gathering the attributes from the remote sensor.
        """

    @abstractmethod
    def disconnect(self, sensor: T) -> None:
        """Disconnect from the provided sensor.

        This removes previously created internal connection objects.

        Arguments:
            sensor (T) The sensor which is supposed to be disconnected.

        Raises:
            SensorNotFound if the sensor does not exist.

            SensorStateException if the sensor's state is not CONNECTED.
        """
        ...

    @abstractmethod
    def start(self, sensor: T) -> None:
        """Start the data acquisition from the provided sensor.

        Arguments:
            sensor (T): the camera that is supposed to be started.

        Raises:
            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """

    @abstractmethod
    def stop(self, sensor: T) -> None:
        """Stop the data acquisition from the provided sensor.

        Arguments:
            sensor (T): the sensor that is supposed to be stopped.

        Raises:
            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not SensorState.STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.
        """

    def reset(self, sensor: T) -> None:
        """Reset the provided sensor.

        A reset
        1. stops image acquisition if image acquisition was started,
        2. performs a sensor reset,
        3. disconnects form the sensor,
        4. connects to the sensor, and
        5. optionally restarts the sensor if image acquisition was started.

        Arguments:
            sensor (T): the sensor that is supposed to be reset.

        Raises:
            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureUnavailable if the feature is (temporarily) not available or not implemented.
        """
        raise FeatureNotImplemented

    @abstractmethod
    def refresh_attributes(self, sensor: T) -> None:
        """Refresh all available attributes.

        Arguments:
            sensor (T): the sensor for which the attributes should be refreshed

        Raises:
            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

        """

    @abstractmethod
    def set_attributes(self, sensor: T) -> None:
        """Apply all attributes that were provided.

        Fails when the first exception is thrown.

        Arguments:
            sensor (T): a sensor configuration that should be applied, that has all the fields.

        Raises:
            DriverStateException: if the driver is in an invalid state

            SensorNotFound:
                If the sensor is currently not found.
            SensorStateException:
                If the sensor's state is not STARTED or CONNECTED.
            SensorStateInconsistent:
                If sensor connection objects are missing even though the state indicates they should be present.
            FeatureNotImplemented:
                If the provided `acquisition_mode` is not implemented in the `_apply_acquisition_mode` method.
        """

    #
    # Sensor Configuration
    #

    def export_configuration(self, sensor: T) -> bytes:
        """Dump the configuration to a bytes array.

        Arguments:
            sensor (T) The sensor which is supposed to be disconnected.

        Return:
            bytes: Configuration dump.

        Raises:
            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotAvailable if the feature is not available.
        """
        raise FeatureNotImplemented

    def import_configuration(self, sensor: T, configuration: bytes) -> None:
        """Read the configuration from a previous configuration dump.

        Arguments:
            sensor (T) The sensor which is supposed to be disconnected.

            configuration (bytes) : Configuration dump.

        Raises:
            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

            FeatureNotAvailable if the feature is not available.
        """
        raise FeatureNotImplemented

    #
    # Callbacks
    #

    def add_listener(self, listener: SensorCallback[T]):
        self.__listeners.append(listener)

    def on_sensor_tracked(self, sensor_identifier: str) -> None:
        for listener in self.__listeners:
            listener.on_sensor_tracked(sensor_identifier)

    def on_sensor_found(self, sensor: T) -> None:
        for listener in self.__listeners:
            listener.on_sensor_found(sensor)

    def on_sensor_lost(self, sensor: T) -> None:
        for listener in self.__listeners:
            listener.on_sensor_lost(sensor)

    def on_sensor_connected(self, sensor: T) -> None:
        for listener in self.__listeners:
            listener.on_sensor_connected(sensor)

    def on_sensor_disconnected(self, sensor: T) -> None:
        for listener in self.__listeners:
            listener.on_sensor_disconnected(sensor)

    def on_sensor_started(self, sensor: T) -> None:
        for listener in self.__listeners:
            listener.on_sensor_started(sensor)

    def on_sensor_stopped(self, sensor: T) -> None:
        for listener in self.__listeners:
            listener.on_sensor_stopped(sensor)

    def on_sensor_updated(self, sensor: T) -> None:
        for listener in self.__listeners:
            listener.on_sensor_updated(sensor)

    #
    # Sensor acquisition
    #

    @abstractmethod
    def execute_trigger(self, sensor: T, timestamp: Optional[int] = None) -> None:
        """Execute a trigger for acquiring data.

        Arguments:
            sensor (T): The camera that is about to be triggered.

        Raises:
            AccessException if acquisition status cannot be reported.

            AttributeError if acquisition status cannot be reported.

            FeatureUnavailable: if the feature is not available.

            InvalidArgumentException: Thrown if it is currently impossible to execute the request, for instance,
                                      because the sensor is not supporting trigger-based acquisition.
        """

    @abstractmethod
    def fetch_sensor_data(self, sensor: T, timeout: float = 0.0) -> Union[SensorData, NotAvailable]:
        ...

    @abstractmethod
    def create_item(self, sensor: T, sensor_data: SensorData):
        ...

    def _sensor(self, sensor: T):
        try:
            return self.sensors[sensor.identifier]
        except KeyError:
            raise SensorNotFound

    def ping(self) -> None:
        """
        If necessary, this can be implemented to check if the connection is still alive.
        This will be executed at a higher frequency, so don't do computationally expensive stuff here.

        Raise if a sensor disconnected.

        If this raises, the node will reset the connection.

        Raises:
            SensorNotFound if the sensor is currently not found.

            SensorStateException if the sensor's state is not CONNECTED or STARTED.

            SensorStateInconsistent if sensor connection objects are missing even though the state indicates they
                should be present.

        """
        pass
