from abc import ABCMeta
from dataclasses import dataclass
from functools import partial

from typing import Generic, TypeVar, List, Optional, Union

from dataspree.inference.core.decorators import action
from dataspree.inference.core.producer import Producer, Property, PropertyType, Item
from dataspree.inference.sensor.data import Camera
from dataspree.inference.sensor.data.camera import TriggerAcquisitionMode, Roi, TimedExposure, ManualGain
from dataspree.inference.sensor.driver import CameraDriver
from dataspree.inference.sensor.node.sensor_input import SensorInput, TemporaryInformationSensor

J = TypeVar('J', bound=Camera)
T = TypeVar('T', bound=CameraDriver)


@dataclass
class TemporaryInformationCamera(TemporaryInformationSensor):
    ...


U = TypeVar('U', bound=TemporaryInformationCamera)


class CameraInput(SensorInput, Generic[J, T, U], metaclass=ABCMeta):
    def __init__(self, name: str, driver: T, required_sensors: Optional[List[str]] = None,
                 sensors_auto_acquisition: Optional[List[str]] = None,
                 delay: float = 0.1,
                 configuration_root_dir: Optional[str] = None,
                 auto_connect_all: bool = True) -> None:
        SensorInput.__init__(self, name, driver=driver,
                             producer=self.create_producer(name), required_sensors=required_sensors,
                             delay=delay, configuration_root_dir=configuration_root_dir,
                             auto_connect_all=auto_connect_all)
        self.sensors_auto_acquisition: Optional[List[str]] = sensors_auto_acquisition

    @classmethod
    def create_producer(cls, name: str) -> Producer:
        return Producer(name, [Property('image', PropertyType.IMAGE)])

    def before_image_acquire(self, sensor: J, driver: T, temporary_information: U) -> bool:
        if self.sensors_auto_acquisition is not None and (sensor.identifier in self.sensors_auto_acquisition):
            if isinstance(sensor.acquisition_mode, TriggerAcquisitionMode):
                driver.execute_trigger(sensor)
        return True

    def on_item_acquired(self, sensor: J, driver: T, temporary_information: U, item: Item) -> Optional[Item]:
        self.logger.debug(f'{sensor}: Frame rate: {temporary_information.fps} '
                          f'vs {sensor.resulting_acquisition_rate}.')
        item['image_count'] = temporary_information.image_counter

        return item

    def create_sensor_representation(self, sensor: J) -> U:
        return TemporaryInformationCamera(producer=self.create_producer(f'{self.name}_{sensor.identifier}'))

    @action
    def highlight_rois(self, sensor: Optional[J] = None):
        def executor(connection: T):
            sensors = {sensor.identifier: sensor} if sensor is not None else connection.sensors
            current_sensor: J
            for current_sensor in sensors.values():
                for roi in current_sensor.camera_model.rois:
                    connection.highlight_roi(current_sensor, roi)

        self.schedule_main_loop_execution(executor, None)

    @action
    def acquisition_continuous(self, sensor: Optional[J] = None):
        def executor(connection: T):
            sensors = {sensor.identifier: sensor} if sensor is not None else connection.sensors
            current_sensor: J
            for current_sensor in sensors.values():
                connection.set_acquisition_continuous(current_sensor)

        self.schedule_main_loop_execution(executor, None)

    @action
    def acquisition_triggered(self, sensor: Optional[J] = None):
        def executor(connection: T):
            sensors = {sensor.identifier: sensor} if sensor is not None else connection.sensors
            current_sensor: J
            for current_sensor in sensors.values():
                connection.set_acquisition_trigger_based(current_sensor)

        self.schedule_main_loop_execution(executor, None)

    @action
    def trigger_exposure(self, sensor: Optional[J] = None, start: bool = True):
        def executor(connection: T):
            sensors = {sensor.identifier: sensor} if sensor is not None else connection.sensors
            current_sensor: J
            for current_sensor in sensors.values():
                connection.execute_software_trigger_exposure(current_sensor, start)

        self.schedule_main_loop_execution(executor, None)

    @action
    def trigger_frame_burst(self, sensor: Optional[J] = None, start: bool = True):
        def executor(connection: T):
            sensors = {sensor.identifier: sensor} if sensor is not None else connection.sensors
            current_sensor: J
            for current_sensor in sensors.values():
                connection.execute_software_trigger_frame_burst(current_sensor, start)

        self.schedule_main_loop_execution(executor, None)

    @action
    def exposure_auto(self, sensor: Optional[J] = None):
        def executor(connection: T):
            sensors = {sensor.identifier: sensor} if sensor is not None else connection.sensors
            current_sensor: J
            for current_sensor in sensors.values():
                connection.set_exposure_auto(current_sensor)

        self.schedule_main_loop_execution(executor, None)

    @action
    def exposure_finder(self, sensor: Optional[J] = None,
                        upper_limit: Optional[Union[float, int]] = None,
                        lower_limit: Optional[Union[float, int]] = None,
                        auto_target_brightness: Optional[float] = None,
                        roi_configuration: Optional[Union[str, Roi]] = None,
                        bounds_check: bool = True):
        def executor(connection: T):
            sensors = {sensor.identifier: sensor} if sensor is not None else connection.sensors
            current_sensor: J
            for current_sensor in sensors.values():
                connection.set_exposure_manual_through_automatic_finder(
                    current_sensor, upper_limit, lower_limit, auto_target_brightness,
                    roi_configuration, bounds_check)

                if isinstance(current_sensor.brightness.exposure, TimedExposure):
                    self.logger.info(f'Found exposure time {current_sensor.brightness.exposure.time}.')

        self.schedule_main_loop_execution(executor, None)

    @action
    def gain_timed(self, sensor: Optional[J] = None, exposure_time: float = 1.0, bounds_check: bool = True):
        def executor(connection: T):
            sensors = {sensor.identifier: sensor} if sensor is not None else connection.sensors
            current_sensor: J
            for current_sensor in sensors.values():
                connection.set_exposure_timed(current_sensor, exposure_time, bounds_check=bounds_check)

                if isinstance(current_sensor.brightness.gain, ManualGain):
                    self.logger.info(f'Set gain to {current_sensor.brightness.gain.value}.')

        self.schedule_main_loop_execution(executor, None)

    @action
    def gain_auto(self, sensor: Optional[J] = None,
                  upper_limit: Optional[Union[float, int]] = None,
                  lower_limit: Optional[Union[float, int]] = None,
                  auto_target_brightness: Optional[float] = None,
                  roi_configuration: Optional[Union[str, Roi]] = None,
                  digital: Optional[bool] = None, bounds_check: bool = True):
        def executor(connection: T):
            sensors = {sensor.identifier: sensor} if sensor is not None else connection.sensors
            current_sensor: J
            for current_sensor in sensors.values():
                connection.set_gain_auto(current_sensor, upper_limit, lower_limit,
                                         auto_target_brightness, roi_configuration,
                                         digital, bounds_check)

        self.schedule_main_loop_execution(executor, None)

    @action
    def gain_finder(
            self, sensor: Optional[J] = None, upper_limit: Optional[Union[float, int]] = None,
            lower_limit: Optional[Union[float, int]] = None,
            auto_target_brightness: Optional[float] = None,
            roi_configuration: Optional[Union[str, Roi]] = None,
            digital: Optional[bool] = None, bounds_check: bool = True):

        def executor(connection: T):
            sensors = {sensor.identifier: sensor} if sensor is not None else connection.sensors
            current_sensor: J
            for current_sensor in sensors.values():
                connection.set_gain_manual_through_automatic_finder(
                    current_sensor, upper_limit, lower_limit, auto_target_brightness,
                    roi_configuration, digital, bounds_check)

                if isinstance(current_sensor.brightness.gain, ManualGain):
                    self.logger.info(f'Set gain {current_sensor.brightness.gain.value} ('
                                     f'digital={current_sensor.brightness.gain.digital})')

        self.schedule_main_loop_execution(executor, None)

    @action
    def white_balance_auto(self, sensor: Optional[J] = None,
                           roi_configuration: Optional[Union[str, Roi]] = None,
                           adjustment_damping: Optional[float] = None,
                           bounds_check: bool = True):
        def executor(connection: T):
            sensors = {sensor.identifier: sensor} if sensor is not None else connection.sensors
            current_sensor: J
            for current_sensor in sensors.values():
                connection.set_white_balance_auto(current_sensor, roi_configuration,
                                                  adjustment_damping, bounds_check)

        self.schedule_main_loop_execution(executor, None)

    @action
    def white_balance_reset(self, sensor: Optional[J] = None):
        def executor(connection: T):
            sensors = {sensor.identifier: sensor} if sensor is not None else connection.sensors
            current_sensor: J
            for current_sensor in sensors.values():
                connection.set_white_balance_manual(current_sensor)

        self.schedule_main_loop_execution(executor, None)

    @action
    def gain_manual(self, sensor: Optional[J] = None,
                    value: Union[float, int] = 0,
                    digital: Optional[bool] = None,
                    bounds_check: bool = True):
        def executor(connection: T):
            sensors = {sensor.identifier: sensor} if sensor is not None else connection.sensors
            current_sensor: J
            for current_sensor in sensors.values():
                connection.set_gain_manual(current_sensor, value, digital, bounds_check)

        self.schedule_main_loop_execution(executor, None)

    def get_exposure_executor(self, increment=10000.0, increase: bool = True, sensor: Optional[J] = None):
        def executor(connection: T):
            sensors = {sensor.identifier: sensor} if sensor is not None else connection.sensors
            self.logger.info('Setting exposure time.')
            current_sensor: J
            for current_sensor in sensors.values():
                exposure = current_sensor.brightness.exposure
                exposure_time = 0.0 if increase else 1000000.0
                if isinstance(exposure, TimedExposure):
                    exposure_time = exposure.time
                exposure_time += increment if increase else -increment

                connection.set_exposure_timed(current_sensor, exposure_time=exposure_time,
                                              bounds_check=True)

                self.logger.info(f'Set exposure time to {exposure_time}.')

        return executor

    @action
    def exposure_plus(self, sensor: Optional[J] = None):
        self.schedule_main_loop_execution(self.get_exposure_executor(sensor=sensor, increase=True), None)

    @action
    def exposure_minus(self, sensor: Optional[J] = None):
        self.schedule_main_loop_execution(self.get_exposure_executor(sensor=sensor, increase=False), None)

    def get_gain_executor(self, increment=10.0, increase: bool = True, sensor: Optional[J] = None,
                          digital: Optional[bool] = None):
        def executor(connection: T):
            sensors = {sensor.identifier: sensor} if sensor is not None else connection.sensors
            current_sensor: J
            for current_sensor in sensors.values():
                gain = current_sensor.brightness.gain
                manual_gain = 0.0 if increase else 10000.0
                if isinstance(gain, ManualGain):
                    manual_gain = gain.value
                manual_gain += increment if increase else -increment

                connection.set_gain_manual(current_sensor, manual_gain, digital, bounds_check=True)
                self.logger.info(f'Set gain to {manual_gain}.')

        return executor

    @action
    def gain_plus(self, sensor: Optional[J] = None, digital: Optional[bool] = None):
        self.schedule_main_loop_execution(self.get_gain_executor(sensor=sensor, increase=True, digital=digital), None)

    @action
    def gain_minus(self, sensor: Optional[J] = None, digital: Optional[bool] = None):
        self.schedule_main_loop_execution(self.get_gain_executor(sensor=sensor, increase=False, digital=digital), None)

    @action
    def user_set_set(self, sensor: Optional[J] = None, name: Optional[str] = None):
        def executor(connection: T):
            sensors = {sensor.identifier: sensor} if sensor is not None else connection.sensors
            current_sensor: J
            for current_sensor in sensors.values():
                str_name = name
                if str_name is None:
                    try:
                        str_name = current_sensor.camera_model.available_configuration_names_writable[0]
                    except IndexError:
                        self.logger.warning(f'No writable user sets for {current_sensor.identifier}.')
                        continue
                connection.set_current_camera_configuration_name(current_sensor, str_name)
        self.schedule_main_loop_execution(executor, None)

    @action
    def user_set_load(self, sensor: Optional[J] = None):
        def executor(connection: T):
            sensors = {sensor.identifier: sensor} if sensor is not None else connection.sensors
            current_sensor: J
            for current_sensor in sensors.values():
                connection.load_current_camera_configuration(current_sensor)
        self.schedule_main_loop_execution(executor, None)

    @action
    def user_set_save(self, sensor: Optional[J] = None):
        def executor(connection: T):
            sensors = {sensor.identifier: sensor} if sensor is not None else connection.sensors
            current_sensor: J
            for current_sensor in sensors.values():
                connection.save_current_camera_configuration(current_sensor)
        self.schedule_main_loop_execution(executor, None)

    @action
    def user_set_save_as_start(self, sensor: Optional[J] = None):
        def executor(connection: T):
            sensors = {sensor.identifier: sensor} if sensor is not None else connection.sensors
            current_sensor: J
            for current_sensor in sensors.values():
                connection.save_current_configuration_as_startup_configuration(current_sensor)
        self.schedule_main_loop_execution(executor, None)
