from typing import Optional, TypeVar, Generic
from logging import getLogger

from dataspree.inference.sensor.data.camera import Brightness, TimedExposure

from dataspree.inference.core.signals import NodeSignal
from dataspree.inference.core.node import Node
from dataspree.inference.sensor.data import Sensor, Camera
from dataspree.inference.sensor.data.sensor import SensorStateTransition

logger = getLogger(__name__)


class RequestFailedSignal(NodeSignal):
    """Signal that is ment not to be instantiated, but provides a unique interface to all the failed signals. """

    def __init__(self, source: Node, reason: Optional[Exception] = None) -> None:
        NodeSignal.__init__(self, source)
        self.reason = reason

    def set_reason(self, reason: Optional[Exception] = None) -> None:
        self.reason = reason


#
# Driver state
#


class DriverConnectedSignal(NodeSignal):
    def __init__(self, source: Node):
        NodeSignal.__init__(self, source)


class DriverDisconnectedSignal(NodeSignal):
    def __init__(self, source: Node, reason: Optional[Exception] = None) -> None:
        NodeSignal.__init__(self, source)
        self.reason: Optional[Exception] = reason


#
# Request driver reset
#

class DriverRequestReset(NodeSignal):
    def __init__(self, source: Node) -> None:
        NodeSignal.__init__(self, source)


class DriverRequestResetFailed(RequestFailedSignal):
    def __init__(self, source: Node, request: DriverRequestReset,
                 reason: Optional[Exception] = None) -> None:
        RequestFailedSignal.__init__(self, source, reason=reason)
        self.request: SensorRequestReconnectSignal[T] = request


T = TypeVar('T', bound=Sensor)


#
# Report state transition
#


class SensorStateTransitionSignal(NodeSignal, Generic[T]):
    def __init__(self, source: Node, sensor: T, target_state: SensorStateTransition):
        NodeSignal.__init__(self, source)
        self.sensor: T = sensor
        self.target_state_transition: SensorStateTransition = target_state


class CameraStateTransitionSignal(SensorStateTransitionSignal[Camera]):
    ...


#
# Request settings application
#

class SensorApplySettingsRequestSignal(NodeSignal, Generic[T]):
    def __init__(self, source: Node, settings: T) -> None:
        NodeSignal.__init__(self, source)
        self.settings: T = settings


class CameraApplySettingsRequestSignal(SensorApplySettingsRequestSignal[Camera]):
    ...


class SensorApplySettingsRequestFailedSignal(RequestFailedSignal, Generic[T]):
    def __init__(self, source: Node, request: SensorApplySettingsRequestSignal[T],
                 reason: Optional[Exception] = None) -> None:
        RequestFailedSignal.__init__(self, source, reason=reason)
        self.request: SensorApplySettingsRequestSignal[T] = request


class CameraApplySettingsRequestFailedSignal(SensorApplySettingsRequestFailedSignal[Camera]):
    ...


#
# Request trigger
#

class SensorTriggerRequestSignal(NodeSignal, Generic[T]):
    def __init__(self, source: Node, sensor: Optional[T], trigger_all: bool,
                 timestamp: Optional[int], adjust_acquisition_mode: bool = False) -> None:
        NodeSignal.__init__(self, source)
        self.sensor: Optional[T] = sensor
        self.all: bool = trigger_all
        self.timestamp: Optional[int] = timestamp
        self.adjust_acquisition_mode: bool = adjust_acquisition_mode


class CameraTriggerSignal(SensorTriggerRequestSignal[Camera]):
    ...


class SensorTriggerRequestSignalFailed(RequestFailedSignal, Generic[T]):
    def __init__(self, source: Node, request: SensorTriggerRequestSignal[T],
                 reason: Optional[Exception] = None) -> None:
        RequestFailedSignal.__init__(self, source, reason=reason)
        self.request: SensorTriggerRequestSignal[T] = request


class CameraTriggerSignalFailed(SensorTriggerRequestSignalFailed[Camera]):
    ...


def create_exposure_signal(node: Node, exposure_time: float, camera_id: Optional[str],
                           timestamp: Optional[int], adjust_acquisition_mode: bool) -> CameraTriggerSignal:
    """
    Return:
        CameraTriggerSignal

    Args:
        node (Node): the source node
        exposure_time (float) exposure time in microseconds.
        camera_id (Optional[str]) Optional camera identifier. If None, trigger and write all.
    """

    trigger_all = camera_id is None
    sensor: Camera = Camera(camera_id, brightness=Brightness(exposure=TimedExposure(exposure_time)))
    return SensorTriggerRequestSignal(node, sensor, trigger_all=trigger_all, timestamp=timestamp,
                                      adjust_acquisition_mode=adjust_acquisition_mode)


#
# Request (re)connect
#

class SensorRequestReconnectSignal(NodeSignal, Generic[T]):
    def __init__(self, source: Node, sensor: T) -> None:
        NodeSignal.__init__(self, source)
        self.sensor: T = sensor


class CameraRequestReconnectSignal(SensorRequestReconnectSignal[Camera]):
    ...


class SensorRequestReconnectFailedSignal(RequestFailedSignal, Generic[T]):
    def __init__(self, source: Node, request: SensorRequestReconnectSignal[T],
                 reason: Optional[Exception] = None) -> None:
        RequestFailedSignal.__init__(self, source, reason=reason)
        self.request: SensorRequestReconnectSignal[T] = request


class CameraRequestReconnectFailedSignal(SensorRequestReconnectFailedSignal[Camera]):
    ...


#
# Request disconnect
#

class SensorRequestDisconnectSignal(NodeSignal, Generic[T]):
    def __init__(self, source: Node, sensor: T) -> None:
        NodeSignal.__init__(self, source)
        self.sensor: T = sensor


class CameraRequestDisconnectSignal(SensorRequestDisconnectSignal[Camera]):
    ...


class SensorRequestDisconnectFailedSignal(RequestFailedSignal, Generic[T]):
    def __init__(self, source: Node, request: SensorRequestDisconnectSignal[T],
                 reason: Optional[Exception] = None) -> None:
        RequestFailedSignal.__init__(self, source, reason=reason)
        self.request: SensorRequestDisconnectSignal[T] = request


class CameraRequestDisconnectFailedSignal(SensorRequestDisconnectFailedSignal[Camera]):
    ...
